// RiskClient.java

package uk.org.redfelineninja.risk.client;

import uk.org.redfelineninja.risk.client.*;
import uk.org.redfelineninja.risk.data.*;
import uk.org.redfelineninja.risk.net.*;
import uk.org.redfelineninja.awt.*;
import uk.org.redfelineninja.risk.test.*;

import java.io.*;

import javax.swing.*;

/**
 * This class provide a means to the start the client.
 *
 * In addition to bootstrapping the client this class aslo provides
 * communication functions. This class as called by the network sub-system
 * to perform all actions on the client. In addition it provides ActionListeners
 * to communicate user actions to the server.
 * <p>
 * Status: ActionListeners are not yet implemented.
 */
public class RiskClient implements RiskClientInterface {
    protected RiskClientFrame mainFrame;
    protected RiskWaitingFrame waitingFrame;
    protected RiskTalkerFrame talkerFrame;
    protected RiskCardFrame cardFrame;
    protected RiskServerInterface server;

    /**
     * Starting point for the application version.
     *
     * This method is responsible for creating all the major frames for
     * this Risk client.
     */
    public static void main(String[] args) {
        // attempt to set the prefered look and feel
        /*
        try {
            String LnF = UIManager.getSystemLookAndFeelClassName();
            UIManager.setLookAndFeel(LnF);
        } catch (Exception e) {
            // bugger
            System.err.println("Can not set system look and feel");
        }
        */

        // start more than one client if requested
        if (args.length == 1) {
            for (int i = Integer.parseInt(args[0]); i > 0; i--) {
                new RiskClient();
            }
        } else {
            new RiskClient();
        }
    }

    /**
     * Start the application version of the Risk client.
     */
    public RiskClient() {
        server = new RiskServerHandler(this);

        // set up the talker frame
        BasicTalker talker = new BasicTalker();
        talkerFrame = new RiskTalkerFrame(talker);
        talker.addActionListener(new RiskTalkerListener(server));

        // we dont yet need the waiting frame
        waitingFrame = null;

        // this defaults to empty
        cardFrame = new RiskCardFrame();

        // this is the big one
        mainFrame = new RiskClientFrame(this);
        mainFrame.setVisible(true);
    }

    // ACCESSORS

    public RiskClientFrame getMainFrame() {
        return mainFrame;
    }
    public RiskWaitingFrame getWaitingFrame() {
        return waitingFrame;
    }
    public RiskTalkerFrame getTalkerFrame() {
        return talkerFrame;
    }
    public RiskCardFrame getCardFrame() {
        return cardFrame;
    }
    public RiskServerInterface getServer() {
        return server;
    }

    // SERVER ACTIONS

    public void gameStarting(Player[] players) {
        waitingFrame = new RiskWaitingFrame(players);
    }

    public void newPlayer(Player player) {
        if (waitingFrame == null) {
            Player[] p = { player };
            waitingFrame = new RiskWaitingFrame(p);
        }

        waitingFrame.newPlayer(player);
    }

    /**
     * Spectator interaction is not yet supported. This function is provided
     * only for compatability.
     */
    public void newSpectator(Player player) {}

    /**
     * Player loss is handled purely at the server. This function is provided
     * for compatability.
     */
    public void lostPlayer(Player player) {}

    public void newGame(Player[] players) {
        // hide and forget the waitingFrame
        if (waitingFrame != null) {
            waitingFrame.setVisible(false);
            waitingFrame = null;
        }

        // register with the turnbar
        mainFrame.getTurnBar().setPlayers(players);
    }

    public void gameOver(Player player) {
        new RiskVictoryFrame(player);
    }

    public void gameOver(Player player, String reason) {
        String msg = (player.name + " halted the game because " + reason);
        JOptionPane.showMessageDialog(null, msg, "Game Over", JOptionPane.WARNING_MESSAGE);
    }

    // TALKER STUFF

    public void message(String message) {
        Writer w = talkerFrame.getTalker().getWriter();

        try {
            w.write(message + '\n');
            w.flush();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void message(String message, Player player) {
        Writer w = talkerFrame.getTalker().getWriter();

        try {
            w.write("Secret message from " + player.name + '\n');
            w.write(message + '\n');
            w.flush();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    // DISPLAY MAINTAINANCE

    /**
     * Moves the turn bar so that the supplied player is at the top
     */
    public void setTurn(Player player) {
        // tell the turn bar
        mainFrame.getTurnBar().turn(player);
    }

    public void setArmies(String country, int a) {
        if (country.equals("status")) {
            mainFrame.getStatusBar().setArmies(a);
        } else {
            mainFrame.getRiskMap().setArmies(country, a);
        }
    }

    public void setOwner(String country, Player player) {
        RiskMap m = mainFrame.getRiskMap();

        m.setColor(player.color);
        m.paintZone(country);
    }

    public void select(String country) {
        mainFrame.getRiskMap().setSelection(country);
    }

    public void showStatus(String status) {
        mainFrame.getStatusBar().setStatus(status);
    }

    /**
     * Die displays are not yet implemented in this interface. This function is
     * provided only for compatibilty.
     */
    public void setDice(int[] die) {
        StringBuffer sb = new StringBuffer("Dice:");

        for (int i = 0; i < die.length; i++) {
            if (die[i] != 0)
                sb.append(" ").append(die[i]);
        } // for( i )

        showStatus(sb.toString());
    }

    public void setCards(Card[] cards) {
        cardFrame.setCards(cards);
    }

} // RiskClient.class
