// RiskServerHandler.java

package uk.org.redfelineninja.risk.client;

import uk.org.redfelineninja.risk.data.Player;
import uk.org.redfelineninja.risk.data.Card;

import uk.org.redfelineninja.risk.net.RiskServerInterface;
import uk.org.redfelineninja.risk.net.RiskClientInterface;
import uk.org.redfelineninja.risk.net.Protocol;

import uk.org.redfelineninja.util.CommandTokenizer;

import java.io.*;
import java.net.*;
import java.util.Hashtable;
import java.util.Vector;

/**
 * This class provides all the tools for a client to get data to and from the
 * Server.
 */
public class RiskServerHandler implements RiskServerInterface, Runnable {

    // the client on which actions will be performed.
    private RiskClientInterface client;

    // store the connection status
    private boolean connected;

    // the connection
    Socket connection;

    // the command string (JDK1.1 (Linux) appears to have bugs that cause chars
    // to come out as ints in strings if care is not taken)
    private static final String cmdStr = String.valueOf(Protocol.command);

    // input and output streams
    private PrintWriter output = null;
    private CommandTokenizer input = null;

    // player registry
    private Hashtable playerRegistry = new Hashtable(); // (String -> Player)

    /**
     * Create a new RiskServerHandler.
     */
    public RiskServerHandler(RiskClientInterface inter) {
        client = inter;
        connected = false;
    }

    // OUTPUT HANDLING

    /**
     * Open a connection to the server and identify yourself as player.
     */
    public void openConnection(String hostname, int port, Player player)
        throws IOException // IOException is a super class of network exceptions
    {
        // check we are not already connected
        if (connected) {
            throw new ConnectException("Attempted to connect before disconnection");
        }

        // make the connection
        connection = new Socket(hostname, port);
        connected = true;

        // input is a command tokenizer
        input = new CommandTokenizer(new InputStreamReader(connection.getInputStream()));
        input.ordinaryChar('#');
        input.eolIsSignificant(true);

        // output is an auto flushed PrintWriter
        output = new PrintWriter(connection.getOutputStream(), true);

        // identify as player
        output.println(cmdStr + Protocol.sIdentify + " \"" + player.name + "\" " + player.color.getRGB());

        // start the run method as a daemon thread
        Thread session = new Thread(this);
        session.setDaemon(true);
        session.start();
    } // openConnection( String, int, Player )

    /**
     * Close the current connection. This should be considered ill mannered and
     * loutish if a game is in progress,
     */
    public void closeConnection() throws IOException {
        if (connected) {
            connected = false;
            connection.close();
        }
    }

    /**
     * Test whether or not we are connected to the server.
     */
    public boolean isConnected() {
        return connected;
    }

    /**
     * Start a new game. In single game servers this is ignored since the server
     * IS the game.
     */
    public void newGame(String name, boolean random, boolean privacy) {
        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        StringBuffer cmd = new StringBuffer();

        // bulk of command
        cmd.append(cmdStr).append(Protocol.sNewGame).append(" \"").append(name).append("\"");

        if (random)
            cmd.append(" ").append(Protocol.aRandom);

        if (privacy)
            cmd.append(" ").append(Protocol.aPrivacy);

        output.println(cmd.toString());
    } // newGame( ... )

    /**
     * Join an existing game. On single game servers this will be ignored since
     * joining the server is joining the game.
     */
    public void joinGame(String name) {
        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        output.println(cmdStr + Protocol.sJoinGame + " \"" + name + "\"");
    } // joinGame( ... )

    /**
     * Stop waiting for new players and start playing the game.
     */
    public void startGame() {
        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        output.println(cmdStr + Protocol.sStartGame);
    } // startGame()

    /**
     * Halt the current game. This should only be performed in exceptional
     * situations.
     */
    public void stopGame(String reason) {
        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        output.println(cmdStr + Protocol.sStopGame + " \"" + reason + "\"");
    } // stopGame( String )

    /**
     * Send a mesage to all parties.
     */
    public void message(String msg) {
        try {
            // loopback self messages (logging/testing purposes)
            if (msg.substring(0, 2).equals(Protocol.sPrompt)) {
                client.message(msg);
            }
        } catch (StringIndexOutOfBoundsException e) {
            // do nothing (not a self message
        }

        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        output.println(msg);
    } // message( String )

    /**
     * Send a message to the supplied player.
     */
    public void message(String message, Player player) {
        client.message("Private messaging it not yet implemented");
    } // message( ... )

    /**
     * Tell the server that the client has clicked on a country.
     */
    public void countryClicked(String country) {
        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        output.println(cmdStr + Protocol.sCountryClicked + " \"" + country + "\"");
    } // countryClicked( String )

    /**
     * Turn the the supplied cards and exchange them for armies.
     */
    public void cards(Card[] cards) {
        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        StringBuffer cmd = new StringBuffer();

        // bulk of command
        cmd.append(cmdStr).append(Protocol.sCards);

        // add the cards
        for (int i = 0; i < cards.length; i++) {
            cmd.append(" \"" + cards[i].country + "\"");
        } // for

        output.println(cmd);
    } // cards( Card [] );

    /**
     * Tell the server that the client wants to fortify.
     */
    public void fortify() {
        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        output.println(cmdStr + Protocol.sFortify);
    } // fortify()

    /**
     * The the server that the client wants to finish their go.
     */
    public void finishTurn() {
        // are we connected
        if (!connected) {
            showStatus(Words.notConnected);
            return;
        }

        output.println(cmdStr + Protocol.sFinishTurn);
    }

    // INPUT HANDLING

    /**
     * This method runs in a daemon Thread parsing data from the server.
     */
    public void run() {
        if (!connected) {
            return;
        }

        try {
            main_session : do {
                try {
                    // get the first two characters
                    Character a = input.getCharacter();
                    Character b = input.getCharacter();

                    // idenity the command
                    if ((a == null) || (b == null)) {
                        // this is a line of chat (the server will not send blank lines)
                        client.message(input.getLine());
                    } else if (!a.equals(b)) {
                        // this is also a line of chat
                        client.message(input.getLine());
                    } else {
                        // this is a message from the server
                        Character type = input.getCharacter();

                        // find the class of message
                        if (type == null) {
                            this.showStatus(input.getLine());
                        } else {
                            // still need to divide further
                            switch (type.charValue()) {
                                case Protocol.warning :
                                case Protocol.query :
                                    // this is a message for the human user
                                    // (advanced clients can intercept the statusbar
                                    // if required)
                                    this.showStatus(input.getLine());
                                    break;

                                case Protocol.comment :
                                    // this a comment (unusual)
                                    client.message(input.getLine());
                                    break;

                                case Protocol.command :
                                case Protocol.chatCommand :
                                    // this is command that must be parsed
                                    handleClientCommand();
                                    break;

                            } // switch
                        } // finding the class of the server message
                    } // determine the overall type of the command

                    // wind off any further tokens
                    while ((input.ttype != input.TT_EOL) && (input.ttype != input.TT_EOF)) {
                        input.nextToken();
                    }
                } catch (NullPointerException e) {
                    System.err.println(
                        "Caught internal client error. The server probably sent an invalid command\n\nPlease mail the following to D.Thompson@gmx.net and include Risk in the\nsubject line.\n");
                    e.printStackTrace();
                }

            }
            while (!input.getEOF());
        } catch (IOException e) {
            message("Connection from server closed abnormally.");
        }

        // permit further connections
        connected = false;

        // protect against silent failures
        showStatus("Connection from server closed.");
        // protect against being called twice on same object
    }

    /**
     * Show the supplied string on the clients status bar starting from the first
     * letter or digit character. In addition to showing the command on the
     * status bar the message is also send the more permenent talker.
     */
    private void showStatus(String msg) {
        // test for null
        if (msg == null) {
            return;
        }

        // send to talker for permanent record
        client.message(msg);

        // the index of the first alphabetic character
        int index;

        for (index = 0; index < msg.length(); index++) {
            if (Character.isLetterOrDigit(msg.charAt(index))) {
                // this is the start point
                break;
            }
        } // for

        if (index < msg.length()) {
            client.showStatus(msg.substring(index));
        }

    } // showStatus( String )

    /**
     * Treat the remaining input as a command on the client.
     */
    private void handleClientCommand() throws IOException {
        // this method is made very readable because null pointer problems can
        // be ignored (they are caught safely and the error reports are
        // desirable)

        // all commands are a single word
        String cmd = input.getWord();

        if (cmd.equals(Protocol.cNewPlayer)) {
            // get data of the form <str> <int> into a player
            Player player = new Player(input.getText(), input.getNumber().intValue(), true);
            playerRegistry.put(player.name, player);
        } else if (cmd.equals(Protocol.cSetArmies)) {
            // set the number of armies in the country (form <str> <int>)
            client.setArmies(input.getText(), input.getNumber().intValue());
        } else if (cmd.equals(Protocol.cSetOwner)) {
            // form <str> <str>, must look up the player info.
            String country = input.getText();
            Player player = (Player) playerRegistry.get(input.getText());
            client.setOwner(country, player);
        } else if (cmd.equals(Protocol.cSetTurn)) {
            // form <str>, player must be looked up before passing to client
            client.setTurn((Player) playerRegistry.get(input.getText()));
        } else if (cmd.equals(Protocol.cSelect)) {
            client.select(input.getText());
        } else if (cmd.equals(Protocol.cNewGame)) {
            // form <str> ..., each must be looked up
            doNewGame();
        } else if (cmd.equals(Protocol.cGameOver)) {
            // form <player> or <player> <str>
            doGameOver();
        } else if (cmd.equals(Protocol.cSetDice)) {
            // form <int> <int> .. .. ..
            doDice();
        } else {
            client.message("Error: server sent unknown command: " + input.getLine());
        }
    } // handleClientCommand()

    /**
     * Treat the remaining input as a command to start a new game.
     */
    private void doNewGame() throws IOException {
        Vector playersV = new Vector();

        String name = input.getText();
        while (name != null) {
            Player p = (Player) playerRegistry.get(name);

            if (p != null)
                playersV.addElement(p);

            name = input.getText();
        }

        Player[] playersA = new Player[playersV.size()];
        playersV.copyInto(playersA);

        client.newGame(playersA);
    } // doNewGame()

    /**
     * Treat the remaining input as a command for finish a game.
     */
    private void doGameOver() throws IOException {
        Player p = (Player) playerRegistry.get(input.getText());
        String excuse = input.getText();

        if (excuse == null) {
            // the game was won by p
            client.gameOver(p);
        } else {
            // the game was ended by p
            client.gameOver(p, excuse);
        }
    } // doGameOver()

    /**
     * Treat the remaining input as dice.
     */
    private void doDice() throws IOException {
        int[] die = new int[5];

        // copy out inputs into an array
        for (int i = 0; i < 5; i++) {
            Double d = input.getNumber();
            die[i] = (d == null ? 0 : d.intValue());
        } // for( i )

        client.setDice(die);
    } // doDice()
} // RiskServerHandler.class
