// RiskDisplayInterface.java

package uk.org.redfelineninja.risk.net;

import uk.org.redfelineninja.risk.data.*;

/**
 * This interface provides a common means to control any Risk client whether it
 * be the graphical client or a robot player.
 *
 * Only asyncronous action are listed in this interface. Information that 
 * would be better modeled as methods return values are modeled as such and
 * this is handled by the Network sub-system.
 *
 * @author Daniel Thompson, Jan 1999
 */
public interface RiskClientInterface {

    // PLAYER REGISTRY

    /**
     * Called when a new player has joined the game the client is waiting to play
     */
    public void newPlayer(Player player);

    /**
     * Called whenever a new spectator has joined the game.
     *
     * This can be ignored unless the client has a talker that supports private
     * talking. Robots could communicate their internal state like this if
     * desired.
     */
    public void newSpectator(Player player);

    /**
     * Called when a player leaves the game (or if his connection is timed out).
     *
     * Leaving mid-game is anti-social but it will happen.
     */
    public void lostPlayer(Player player);

    // GAME CONTROL

    /**
     * Called when the games owner decides to stop waiting for new players and
     * start playing.
     */
    public void newGame(Player[] players);

    /**
     * Called when the game is won.
     */
    public void gameOver(Player victor);

    /**
     * Called when the game is stopped by the server.
     */
    public void gameOver(Player player, String reason);

    // MESSAGING

    /**
     * Message sent over the chat system to all clients.
     */
    public void message(String message);

    // TURN CONTROL

    /**
     * It is player's turn
     */
    public void setTurn(Player player);

    // DISPLAY CONTROL

    /**
     * Display the supplied string in the clients status line.
     */
    public void showStatus(String status);

    /**
     * Set the number of armies in a country.
     */
    public void setArmies(String country, int a);

    /**
     * Set the owner of the country to the player.
     */
    public void setOwner(String country, Player player);

    /**
     * Select the supplied country.
     */
    public void select(String country);
    /**
     * Supply the current state of the dice to those it interests.
     */
    public void setDice(int[] die);

    /**
     * Tell the client the cards that the player has.
     *
     * Clients can connect to the server with two connections if there are two
     * players at that client.
     */
    public void setCards(Card[] cards);
} // interface
