// GameState.java

package uk.org.redfelineninja.risk.robot;

import uk.org.redfelineninja.risk.client.RiskStatusBar;
import uk.org.redfelineninja.risk.data.*;
import uk.org.redfelineninja.risk.io.MapDefinition;
import uk.org.redfelineninja.risk.net.*;
import uk.org.redfelineninja.risk.server.RiskEngine;

import uk.org.redfelineninja.util.Adjacency;
import uk.org.redfelineninja.util.CommandTokenizer;

import java.io.*;
import java.util.*;

/**
 * This class stores all information about the game state for interagation by
 * the robot. It monitors network messages in order to maintain its own state
 * without any invention from a robot's programmer. From this network activity
 * this class is also responsible for calling the robots makeMove method.
 *
 * <b>A robots author should only use the interogation methods. Those that
 * reflect game state, not that alter game state.</b> Currently GameState
 * allows programmers to access all the methods in the RiskClientInterface.
 * This will be changed for the next version, they will be removed and put
 * somewhere private. Untill then a robot author who calls these
 * funtions is likely to find their robot doing some very strange things.
 */
public class GameState implements RiskClientInterface {
    /**
     * Mode constant, game is in null state.
     *
     * The game should never be in a null state by the time the robot is
     * told to make a move.
     */
    public static final int NULL = 0;

    /**
     * Mode constant, initial selection of countries.
     *
     * In this mode the robot must select any unoccupied country in order to
     * claim it as his own.
     */
    public static final int INIT_SELECTION = RiskEngine.INIT_SELECTION;

    /**
     * Mode constant, placement of initial armies.
     *
     * In this mode the robot must select which country it wishes to place an
     * army into. No attacks will be made following this placement.
     */
    public static final int INIT_PLACEMENT = RiskEngine.INIT_PLACEMENT;

    /**
     * Mode constant, unrestricted placement of armies from the pool.
     *
     * In this mode the robot must select which country it wishes to place an
     * army into. Following this robot may enter SELECTION mode.
     */
    public static final int PLACEMENT = RiskEngine.PLACEMENT;

    /**
     * Mode constant, a friendly country has been selected, we are expecting
     * attack something.
     *
     * Here a robot must pick the country that neighbours the current selection
     * that it wants to attack or pick select to return to selection mode.
     */
    public static final int ATTACK = RiskEngine.ATTACK;

    /**
     * Mode constant, placement restricted after a country has been defeated.
     *
     * In this mode armies may only be placed into the selected country and the
     * country that was defeated. It is a good idea for the robot to record the
     * defeated country, it is currently not stored within the GameState.
     */
    public static final int LIMITED_PLACEMENT = RiskEngine.LIMITED_PLACEMENT;

    /**
     * Mode constant, placement limieted by the rules of fortification.
     *
     * Here the robot must pick countries that are adjacent to the selected
     * country.
     */
    public static final int FORTIFY_PLACEMENT = RiskEngine.FORTIFY_PLACEMENT;

    /**
     * Mode constant, waiting for a country to fortify.
     *
     * The robot must pick any country whose armies need to me moved somewhere
     * else. The robot may also pick done when it has finished.
     */
    public static final int FORTIFY = RiskEngine.FORTIFY;

    /**
     * Mode constant, No country is selected.
     *
     * The robot is required to pick a country to attack with. It may also pick
     * done or fortify.
     */
    public static final int SELECTION = -10;

    private Hashtable index; // country.name -> country.index
    private CountryVector board;
    private Country selection, lastClickedCountry;
    private int pool;

    private int mode = 0;

    private Player me;
    private Player active;
    private Vector playerRegistry; // String -> Player

    private Robot robot;

    // this move is to prevent occilatory behaviour
    private Move hitList;

    private RiskServerInterface server;

    /**
     * Create a new GameState. This takes a map definition file and a Robot.
     * Make sure that the map definition is the same as the one used by the
     * server.
     */
    public GameState(File f, Robot r) throws FileNotFoundException {
        robot = r;
        me = r.getIdentity();

        // load the map definition
        MapDefinition md = new MapDefinition(f);

        Adjacency adj = md.getAdjacency();

        // for a list of labels
        String[] labels = new String[adj.size()];
        adj.copyLabelsInto(labels);

        // make board vector and the index
        board = new CountryVector();
        board.setSize(labels.length);
        index = new Hashtable(labels.length);

        // copy the data in
        for (int i = 0; i < labels.length; i++) {
            board.setElementAt(new Country(labels[i], i), i);
            index.put(labels[i], new Integer(i));
        } // for( i )

        // add in the adjacency information
        for (int i = 0; i < labels.length; i++) {
            // fetch this in the outer loop
            Country row = board.countryAt(i);

            for (int j = 0; j < labels.length; j++) {
                if (adj.linked(labels[i], labels[j])) {
                    row.neighbors.addElement(board.countryAt(j));
                } // linked
            } // for( j )
        } // for( i )

        // make the initial hit list
        hitList = new Move(board.size()).add(1);

    } // contructor( File )

    // INTEROGATION

    /**
     * Get the current mode. This can be used to determine the current game
     * state. Used in conjuction with the modal constants such as PLACEMENT.
     *
     * @return the current mode
     */
    public int getMode() {
        return mode;
    } // getMode()

    /**
     * Test whether the robot associated with this GameState is the active
     * player.
     *
     * @return true of the robot is that active player.
     */
    public boolean myTurn() {
        return active.equals(me);
    } // isActive()

    /**
     * Get the size of the board.
     *
     * @return the number of countries in the board
     */
    public int getBoardSize() {
        return board.size();
    } // getBoardSize()

    /**
     * Get the requested country.
     * @param i index of the country
     * @return the country at the specified index
     */
    public final Country getCountry(int i) {
        return board.countryAt(i);
    } // getCountry()

    /**
     * Get the requested country.
     * @param name the name of the country to be fetched
     * @return the country of the specified 
     */
    public final Country getCountry(String name) {
        try {
            return board.countryAt(((Integer) index.get(name)).intValue());
        } catch (Exception e) {}

        return null;
    }

    /**
     * Get the currently selected country.
     */
    public final Country getSelection() {
        return selection;
    }

    /**
     * Get the absolute strength of each country on the board.
     * <p>
     * Absolute strength is the number of armies each contry contains.
     *
     * @return a Move where each element is the strength of the corresponding
     * country
     */
    public final Move getStrength() {
        Move m = getLegalMoves();

        for (int i = 0; i < m.countries.length; i++) {
            if (m.countries[i] > 0.5) {
                m.countries[i] = (float) board.countryAt(i).armies;
            }
        } // for( i )

        return m;
    } // getStrength()

    /**
     * Get the relative strength of each country on the board.
     * <p>
     * Relative strength is the absolute strength of a country and its
     * friendly neighbours minus the absolute strength of all hostile
     * neighbours.
     *
     * @return a Move where each element is the relative strength of the
     * corresponding country
     */
    public final Move getRelativeStrength() {
        Move m = getLegalMoves();

        for (int i = 0; i < m.countries.length; i++) {
            if (m.countries[i] > 0.5) {
                m.countries[i] = getRelativeStrength(board.countryAt(i));
            }
        } // for( i )

        return m;
    } // getRelativeStrength()

    /**
     * Get the relative strength of a single country.
     * <p>
     * Relative strength is the absolute strength of a country and its
     * friendly neighbours minus the absolute strength of all hostile
     * neighbours.
     *
     * @return the relative strength of the supplied country
     */
    public final float getRelativeStrength(Country c) {
        float strength = c.armies;

        for (int j = 0; j < c.neighbors.size(); j++) {
            Country d = c.neighbors.countryAt(j);
            strength += (d.owner.equals(c.owner) ? d.armies : -d.armies);
        } // for( j )

        return strength;
    } // getRelativeStrength( Country )

    /**
     * Get the percentage ownership of each continent.
     * <p>
     * <strong>This does not work and exists only as a test stub. Do not use this
     * unless you are developing the Risk Arena.</strong> 
     *
     * @return a Move where each country that the player owns has a fraction
     * associated with it. This fraction is the fraction of the conitnent that
     * this player owns.
     */
    public final Move getContinentWeights() {
        return getStrength();
    } // getContinentWeights()

    /**
     * Get the Moves available to the robot.
     * <p>
     * This method is used internally to ensure that the robot only makes legal
     * moves. It can be used by advanced robots to prevent wasting time analysing
     * illegal moves.
     *
     * @return a Move where each element is either 0 or 1. 0 represents an
     * illegal move.
     */
    public Move getLegalMoves() {
        Move m = new Move(board.size());
        float OK = 1;
        Country c;
        int ind;

        switch (mode) {
            case INIT_SELECTION :
                // valid if country.owner is null

                for (int i = 0; i < board.size(); i++) {
                    if (board.countryAt(i).owner == null)
                        m.countries[i] = OK;
                } // for( i )

                break;

            case INIT_PLACEMENT :
            case PLACEMENT :
                // valid if country.owner is me

                for (int i = 0; i < board.size(); i++) {
                    if (board.countryAt(i).owner.equals(me))
                        m.countries[i] = OK;
                } // for( i )

                break;

            case ATTACK :
                // valid for countries adjacent to the selection and not owned
                // by me

                for (int i = 0; i < selection.neighbors.size(); i++) {
                    c = selection.neighbors.countryAt(i);
                    if (!c.owner.equals(me)) {
                        ind = ((Integer) index.get(c.name)).intValue();
                        m.countries[ind] = OK;
                    }
                }

                m.select = OK;

                break;

            case SELECTION :
                // valid for all countries owned by me that are adjacent to a
                // country not owned by me and have armies in

                for (int i = 0; i < board.size(); i++) {
                    c = board.countryAt(i);
                    if (c.owner.equals(me) && (c.armies > 0)) {
                        // find if an adjacent country is not owned by me
                        CountryVector adj = board.countryAt(i).neighbors;
                        for (int j = 0; j < adj.size(); j++) {
                            if (!adj.countryAt(j).owner.equals(me)) {
                                m.countries[i] = OK;
                                break;
                            } // if the adjacent country is owned by someone else
                        } // for( j )
                    } // if I own the country
                } // for( i )

                m.fortify = OK;
                m.done = OK;

                // remove any country that has been left
                m.multiplyBy(hitList);

                break;

            case LIMITED_PLACEMENT :
                // valid for the selection and the defeated country

                ind = ((Integer) index.get(selection.name)).intValue();
                m.countries[ind] = OK;

                ind = ((Integer) index.get(lastClickedCountry.name)).intValue();
                m.countries[ind] = OK;

                break;

            case FORTIFY :
                // any country owned by me with armies in (UNSAFE - robot must
                // implement further protection to prevent blocking)

                for (int i = 0; i < board.size(); i++) {
                    c = board.countryAt(i);
                    if (c.owner.equals(me) && (c.armies > 0))
                        m.countries[i] = OK;
                } // for( i )

                m.done = OK;

                break;

            case FORTIFY_PLACEMENT :
                // any country adjacent to the selection and owned my me

                for (int i = 0; i < selection.neighbors.size(); i++) {
                    c = selection.neighbors.countryAt(i);
                    if (c.owner.equals(me)) {
                        ind = ((Integer) index.get(c.name)).intValue();
                        m.countries[ind] = OK;
                    }
                }
                break;

        } // switch on mode

        return m;
    } // getLegalMoves()

    /**
     * Force the robot to make a move.
     */
    protected void makeMove() {
        Move move = robot.bestMove(mode);
        move.slideAboveZero();
        move.multiplyBy(getLegalMoves());

        int m = move.getMove();

        switch (m) {
            case Move.FORTIFY :
                System.out.println("Fortify");
                server.fortify();
                break;

            case Move.DONE :
                System.out.println("Done");
                server.finishTurn();
                hitList = new Move(board.size()).add(1);
                break;

            case Move.SELECT :
                mode = SELECTION;
                hitList.countries[selection.index] = 0;
                makeMove();
                break;

            default :
                Country c = board.countryAt(m);

                if (mode != LIMITED_PLACEMENT)
                    lastClickedCountry = c;

                switch (mode) {
                    case INIT_SELECTION :
                        System.out.println("Occupying " + c.name);
                        break;

                    case INIT_PLACEMENT :
                    case PLACEMENT :
                    case LIMITED_PLACEMENT :
                        System.out.println("Place armies in " + c.name);
                        break;

                    case ATTACK :
                        System.out.println("Attack " + c.name);
                        break;

                    case SELECTION :
                        System.out.println("Select " + c.name);
                        break;

                    case FORTIFY :
                        System.out.println("Fortify from " + c.name);
                        break;

                    case FORTIFY_PLACEMENT :
                        System.out.println("  to " + c.name);
                        break;

                    default :
                        System.out.println("Click " + c.name);
                        break;
                }
                server.countryClicked(c.name);
        } // switch on m
    } // makeMove()

    /**
     * Set the server that this game state is attached to.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void setServer(RiskServerInterface s) {
        server = s;
    } // setServer()

    // RISK CLIENT INTERFACE

    // PLAYER REGISTRY

    /**
     * Called when a new player has joined the game the client is waiting to play
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void newPlayer(Player player) {
        playerRegistry.addElement(player);
    } // newPlayer

    /**
     * Called whenever a new spectator has joined the game.
     *
     * This can be ignored unless the client has a talker that supports private
     * talking. Robots could communicate their internal state like this if
     * desired.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void newSpectator(Player player) {
        // this can be ignored
    } // newSpectator( Player )

    /**
     * Called when a player leaves the game (or if his connection is timed out).
     *
     * Leaving mid-game is anti-social but it will happen.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void lostPlayer(Player player) {
        // this too can be ignored
    } // lostPlayer( Player )

    // GAME CONTROL

    /**
     * Called when the games owner decides to stop waiting for new players and
     * start playing.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void newGame(Player[] players) {
        // this too can be ignored
    } // newGame( Player )

    /**
     * Called when the game is won.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void gameOver(Player victor) {
        System.out.println(victor.name + " won.");

        try {
            server.closeConnection();
        } catch (IOException e) {
            e.printStackTrace();
        }

        System.exit(0);
    } // gameOver( Player )

    /**
     * Called when the game is stopped by the server.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void gameOver(Player player, String reason) {
        System.out.println(player.name + " ended the game: " + reason);

        try {
            server.closeConnection();
        } catch (IOException e) {
            e.printStackTrace();
        }

        System.exit(0);
    } // gameOver( Player, String )

    // MESSAGING

    /**
     * Message sent over the chat system to all clients.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void message(String message) {
        if (message == null)
            return;

        // THIS IS HIGHLY DEPENDANT ON SERVER STRUCTURE
        if (message.length() == 7) {
            if (message.substring(0, 5).equals("--% #")) {
                try {
                    int m = Integer.parseInt(message.substring(5));
                    mode = m;
                    if (active.equals(me)) {
                        makeMove();
                    }
                    return;
                } catch (NumberFormatException e) {}
            }
        }

        // If this is not a hint then display the message
        System.out.println(message);
    } // message( String )

    // TURN CONTROL

    /**
     * It is player's turn
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void setTurn(Player player) {
        active = player;

        if ((mode != NULL) && active.equals(me)) {
            if (mode == INIT_SELECTION) {
                makeMove();
            } else {
                mode = PLACEMENT;
                makeMove();
            }
        }
    } // setTurn( Player )

    // DISPLAY CONTROL

    /**
     * Display the supplied string in the clients status line.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void showStatus(String status) {
        if (status.equals(Protocol.wGameStarted)) {

            mode = INIT_SELECTION;

            if (active.equals(me)) {
                makeMove();
            }
        }
        // do nothing
    } // showStatus( String )

    /**
     * Set the number of armies in a country.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void setArmies(String country, int a) {
        if (country.equals(RiskStatusBar.status)) {
            pool = a;
        } else {
            int i = ((Integer) index.get(country)).intValue();
            board.countryAt(i).armies = a;
        }
    } // setArmies( String, int )

    /**
     * Set the owner of the country to the player.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void setOwner(String country, Player player) {
        int i = ((Integer) index.get(country)).intValue();
        board.countryAt(i).owner = player;
    } // setOwner( String, Player )

    /**
     * Select the supplied country.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void select(String country) {
        if (country == null) {
            selection = null;
            return;
        }

        int i = ((Integer) index.get(country)).intValue();
        selection = board.countryAt(i);
    } // select( String )

    /**
     * Supply the current state of the dice to those it interests.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void setDice(int[] die) {
        // this does not matter a bit
    }

    /**
     * Tell the client the cards that the player has.
     *
     * Clients can connect to the server with two connections if there are two
     * players at that client.
     * <p><strong>Robot authurs should not normally call this method.</strong>
     */
    public void setCards(Card[] cards) {
        // this is not yet implemented
    } // setCards( Card [] )

} // GameState.class
