// Move.java

package uk.org.redfelineninja.risk.robot;

import java.util.Random;

/**
 * This class represents the choices that the robot wishes to make.
 *
 * The Move can be considered as a vector (mathematical sense not
 * <code>java.lang.Vector</code>). There are as many elements to the vector as
 * there are countries on the board. Each element is a <code>float</code>. The
 * larger that value is the better it is to make a move on the associated
 * country. After all legal moves have been discarded then the largest
 * remaining element will be the move that is made.
 * <p>
 * Also included are single <code>floats</code> select, fortify and done.
 * These are also compared to the elements, if any of these are the largest
 * value then their actions are performed. Please read their entries for more
 * information.
 * <p>
 * Move also provides some maths operations for basic manipulation of Moves
 *
 * @author Daniel Thompson, Mar. 1999
 */
public class Move {
    /** This is the constant return be <code>getMove()</code> when the robot has
     *  opted to return to SELECT mode. */
    public static final int SELECT = -1;

    /** This is the constant return be <code>getMove()</code> when the robot has
     *  opted to fortify its position. */
    public static final int FORTIFY = -2;

    /** This is the constant return be <code>getMove()</code> when the robot has
     *  opted to finish its turn. */
    public static final int DONE = -3;

    private static Random random = new Random();

    /** List of desire levels to click on each country */
    public float[] countries;

    /** The desire level to move to select mode.
     *  This is used by a robot in attack mode to indicate that it wants
     *  wants to return to select mode */
    public float select;

    /** The desire level to move to fortify mode.
     *  This is used by a robot that wants to fortify */
    public float fortify;

    /** The desire level of the robot to finish their turn.
     *  That is used by a robot that wants to finish its turn */
    public float done;

    /**
     * Create a new move of length size.
     */
    public Move(int size) {
        countries = new float[size];
        reset();
    }

    /**
     * Get the move that should be made.
     *
     * @return an integer between 0 and the number of countries <b>or</b> a
     * negative number belonging to the set of non country moves, SELECT,
     * FORTIFYor DONE.
     */
    public int getMove() {
        // this will take the index of the best move in countries
        int index = 0;
        float max = 0;

        // pick the most popular country
        for (int i = 0; i < countries.length; i++) {
            if (max < countries[i]) {
                index = i;
                max = countries[i];
            }
        }

        // chech that the other moves are not a better bet
        if (select >= max) {
            index = SELECT;
            max = select;
        }

        if (fortify >= max) {
            index = FORTIFY;
            max = fortify;
        }

        if (done >= max) {
            index = DONE;
            max = done;
        }

        return index;
    } // getMove()

    /**
     * Reset this move back to entirely zeros.
     */
    public void reset() {
        for (int i = 0; i < countries.length; i++) {
            countries[i] = 0;
        } // for( i )

        select = 0;
        fortify = 0;
        done = 0;
    } // reset()

    /**
     * Multiply this move by the supplied move. This can be used to provide extra
     * weightings are to make a legal move filter (multiply by a Move consisting
     * only of zeros and ones).
     */
    public Move multiplyBy(Move m) {
        for (int i = 0; i < countries.length; i++) {
            countries[i] *= m.countries[i];
        } // for( i )

        select *= m.select;
        fortify *= m.fortify;
        done *= m.done;

        return this;
    } // multiplyBy( Move )

    public Move multiplyBy(float f) {
        for (int i = 0; i < countries.length; i++) {
            countries[i] *= f;
        } // for( i )

        select *= f;
        fortify *= f;
        done *= f;

        return this;
    } // multiplyBy( float )

    public Move add(Move m) {
        for (int i = 0; i < countries.length; i++) {
            countries[i] += m.countries[i];
        } // for( i )

        select += m.select;
        fortify += m.fortify;
        done += m.done;

        return this;
    } // add( Move )

    public Move add(float f) {
        for (int i = 0; i < countries.length; i++) {
            countries[i] += f;
        } // for( i )

        select += f;
        fortify += f;
        done += f;

        return this;
    } // addTo( Move )

    /**
     * Randomly scatter the country weights by the supplied factor.
     * The non-country weights remain unaffected.
     *
     * @param f the weights will be scattered by +/- f
     */
    public Move scatter(float f) {
        float m = f + f;

        for (int i = 0; i < countries.length; i++) {
            countries[i] += (random.nextFloat() * m) - f;
        }
        return this;
    } // scatter( float )

    /**
     * Make all weights positive. This should be called before applying a legal
     * move filter.
     */
    public Move slideAboveZero() {
        float m = 0;

        for (int i = 0; i < countries.length; i++) {
            m = Math.min(m, countries[i]);
        }

        m -= 0.1;

        for (int i = 0; i < countries.length; i++) {
            // slide above 0 + 0.1
            countries[i] -= m;
        }

        select -= m;
        fortify -= m;
        done -= m;

        return this;
    } // slideAboveZero()

    /**
     * Return this Move as a String. This is intended to be used mostly for
     * debugging purposes, when the robot is behaving in an unexpected way and
     * it is difficult to see how.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();

        for (int i = 0; i < countries.length; i++) {
            sb.append(countries[i]).append("     ").append(i).append("\n");
        }

        sb.append(select).append("     Select\n");
        sb.append(fortify).append("     Fortify\n");
        sb.append(done).append("     Done\n");

        return sb.toString();
    }
} // Move.class
