// SimpleRobot.java

package uk.org.redfelineninja.risk.robot;

import uk.org.redfelineninja.risk.data.Player;

import java.util.*;

/**
 * This is a simple example robot showing how to write vector players for the
 * Risk Arena.
 */
public class SimpleRobot extends Robot {
    /** Random number generator. */
    protected static Random random = new Random();

    protected int lastMove;

    public SimpleRobot() {
        super();

        defaultName = "Simon";
    }

    public Move bestMove(int mode) {
        Move m;

        switch (mode) {
            case GameState.INIT_PLACEMENT :
                m = bestPlacement();
                m.multiplyBy(bestBorderPlacement());
                return m;

            case GameState.PLACEMENT :
                return bestPlacement();

            case GameState.LIMITED_PLACEMENT :
                lastMove = 0;
                return bestBorderPlacement();

            case GameState.ATTACK :
                fortify = false;
                return bestAttack();

            case GameState.SELECTION :
                return bestSelection();

            case GameState.FORTIFY :
                return bestFortify();

            case GameState.FORTIFY_PLACEMENT :
                return bestBorderPlacement();

            default :
                System.out.println(me.name + " had to resort to a random move.");
                m = new Move(game.getBoardSize());
                m.scatter(1);
                m.select = 0.8f;
                m.done = 0.8f;
                return m;
        } // switch on mode
    } // makeMove( int )

    public static void main(String[] args) {
        Robot r = new SimpleRobot();
        Robot.main(r, args);
    } // main( String [] )

    public Move bestPlacement() {
        Move myMove = game.getStrength();

        myMove.multiplyBy(0.5f);
        myMove.add(game.getRelativeStrength().multiplyBy(-1));
        myMove.scatter(5f);

        myMove.slideAboveZero();
        myMove.add(1);

        // kick out anything that is not bordering another country
        failed : for (int i = 0; i < myMove.countries.length; i++) {
            Country c = game.getCountry(i);
            for (Enumeration e = c.neighbors.elements(); e.hasMoreElements();) {
                if (!((Country) e.nextElement()).owner.equals(me))
                    continue failed;
            } // for( e )

            // make awkward but not impossible
            myMove.countries[i] = 0.1f;
        } // for( i )

        return myMove;
    } // bestPlacement()

    public Move bestSelection() {
        Move myMove = game.getStrength();

        myMove.multiplyBy(0.5f);
        myMove.add(game.getRelativeStrength());
        myMove.scatter(1.5f);

        // remove non optimal attacks
        for (int i = 0; i < myMove.countries.length; i++) {
            if (game.getCountry(i).armies < 3) {
                myMove.countries[i] = 0;
            }
        } // for( i )

        myMove.fortify = 2;

        return myMove;
    } // bestSelect()

    public Move bestAttack() {
        Country c = game.getSelection();

        Move myMove = new Move(game.getBoardSize());

        // if we can not make an optimal attack then make us likely not to
        // select
        myMove.select = 1;
        if (c.armies < 3) {
            return myMove;
        }

        float str = game.getRelativeStrength(c);

        for (Enumeration e = c.neighbors.elements(); e.hasMoreElements();) {
            Country n = (Country) e.nextElement();

            myMove.countries[n.index] = str - game.getRelativeStrength(n);
        } // for( e )

        // make equally good moves random
        myMove.scatter(0.5f);

        return myMove;
    } // bestAttack()

    // provides safety against locked behaviour
    boolean fortify;

    /** Random move */
    public Move bestFortify() {
        Move myMove = new Move(game.getBoardSize());

        // has fortify been called twice in a row
        if (fortify) {
            myMove.done = 10;
            return myMove;
        }

        fortify = true;

        myMove.done = 0.1f;

        // identify any stranded armies
        failed : for (int i = 0; i < myMove.countries.length; i++) {
            Country c = game.getCountry(i);
            if (c.armies > 0) {
                for (Enumeration e = c.neighbors.elements(); e.hasMoreElements();) {
                    if (!((Country) e.nextElement()).owner.equals(me))
                        continue failed;
                } // for( e )

                myMove.countries[i] = (float) Math.random();
            }
        } // for( i )

        lastMove = myMove.getMove();

        return myMove;
    } // bestFortify()

    /**
     * A weighting based on the number of foreign borders.
     */
    public Move bestBorderPlacement() {
        fortify = false;

        Move myMove = new Move(game.getBoardSize());

        for (int i = 0; i < myMove.countries.length; i++) {
            // weight everything above the current country
            if (i != lastMove)
                myMove.countries[i]++;

            Country c = game.getCountry(i);
            for (Enumeration e = c.neighbors.elements(); e.hasMoreElements();) {
                if (!((Country) e.nextElement()).owner.equals(me)) {
                    myMove.countries[i]++;
                }
            } // for( e )
        } // for( i );

        // make countries of equal merit randomly choosen
        myMove.scatter(0.5f);

        return myMove;
    } // bestFortifyPlacement()

} // SimpleRobot.class
