// RiskEngine.java

package uk.org.redfelineninja.risk.server;

import uk.org.redfelineninja.risk.data.*;
import uk.org.redfelineninja.risk.net.*;
import uk.org.redfelineninja.risk.io.MapDefinition;
import uk.org.redfelineninja.risk.client.RiskStatusBar;

import uk.org.redfelineninja.util.*;

import java.util.*;

/**
 * This class is the heart of the Risk system.
 *
 * This is a very large class that provides all the control and state broadcast
 * for the Risk Arena. It is suitable for use in all servers and has been 
 * been extensively tested in use. There has been no boundary testing however
 * so the odd bug may still remain.
 */
public class RiskEngine {
    // THIS CLASS HAS AN UNPLEASANT AMOUNT OF STATE INFORMATION

    // the players playing risk
    private Player[] players;

    // the active player
    private int turn;

    // the adjacency information
    private Adjacency adj;

    // the game state table (String -> Country)
    private Hashtable board;

    // the make up of the land masses
    private Hashtable continents;

    // the number of unassigned countries (INIT_SELECTION mode) / armies (INIT_PL ..)
    private int unassigned;

    // the currently selected country
    private Country selection;

    // the country that was defeated (used in LIMITED_PLACEMENT
    private Country defeated;

    // the number of armies not in a country
    private int armiesInTransit;

    // the dice
    private Dice die;

    // this is the total number of initial armies on the board
    private int initialArmies = 80;
    // this is the rate at which the armies are distributed (small is fastest)
    // in real risk this should be large but that is awkward
    // private float armiesDivisor = 5.0f;
    private float armiesDivisor = 2.0f;

    // the current mode (see mode constants)
    private int mode;

    // the class by which actions are communicated out of the server
    private RiskClientWriter client = null;

    // mode constants
    public static final int INIT_SELECTION = -1;
    public static final int INIT_PLACEMENT = -2;
    public static final int PLACEMENT = -3;
    public static final int ATTACK = -4;
    public static final int LIMITED_PLACEMENT = -5;
    public static final int FORTIFY_PLACEMENT = -6;
    public static final int FORTIFY = -7;

    // status constants
    public static final int OK = 0;
    public static final int NO_ARMIES = 1;
    public static final int NOT_CURRENT_PLAYER = 2;
    public static final int NOT_ADJACENT = 3;
    public static final int NOT_OWNER = 4;
    public static final int ILLEGAL_ACTION = 5;
    public static final int NO_SUCH_COUNTRY = 6;
    public static final int NOT_ALLOWED = 7;
    public static final int GAME_OVER = 8;

    /**
     * Create a new risk engine.
     *
     * @param  p	    array of player to play this game
     * @param  a	    adjacency matrix describing the board
     * @param  ct    hashtable (String -&gt; Continent)
     * @param  clnt  the network interface for this engine
     */
    public RiskEngine(Player[] p, Adjacency a, Hashtable ct, RiskClientWriter clnt) {
        // simple init
        players = p;
        adj = a;
        continents = ct;
        client = clnt;
        turn = 0;
        mode = INIT_SELECTION;

        // set the number of unassigned countries
        unassigned = adj.size();

        // we can now create the countries data structure
        board = new Hashtable();

        // make up the board, cause trouble if the adjacency matrix is invalid
        for (Enumeration e = adj.labels(); e.hasMoreElements();) {
            Object obj = e.nextElement();
            if (obj instanceof String) {
                Country c = new Country((String) obj);

                // add the country to the board
                board.put((String) obj, c);
            } else {
                // this is a serious problem
                throw new InternalServerException("The adjacency matrix contains something that is not java.lang.String");
            }
        } // for

        // make the die
        die = new Dice(client);

        // start the game
        client.newGame(players);
        client.setTurn(players[turn]);
    } // constructor( Player [], Adjacency )

    /**
     * Assign the players their countries randomly.
     */
    public void randomise() {
        int size = adj.size();

        // check for sanity
        if (unassigned != size) {
            throw new InternalServerException("Can't randomise. Some countries have been assigned.");
        }

        // here we create a random number to pick the country and suffle the
        // array down in memory and repeat. Not efficient but it is likely to
        // be uncomplicated

        Object[] labels = new Object[size];
        adj.copyLabelsInto(labels);

        // until all countries have been assigned do
        while (size > 0) {
            // why floor does not return an int is beyond me
            int index = (int) Math.floor(Math.random() * size);

            // send a click to the normal interface, saves rewriting everything
            this.click(labels[index].toString(), players[turn]);

            // shuffle array
            size--;
            for (int i = index; i < size; i++) {
                labels[i] = labels[i + 1];
            }
        } // while
    } // randomise()

    /**
     * Returns the player whose turn it is.
     */
    public synchronized Player getActivePlayer() {
        return players[turn];
    } // getActivePlayer()

    public synchronized boolean liveGame() {
        boolean live = false;

        for (int i = 0; i < players.length; i++) {
            if (players[i].playing) {
                live = true;
                break;
            }
        } // for( i )

        return live;
    } // check that the game is still alive

    /**
     * Returns the current mode.
     */
    public int getMode() {
        return mode;
    } // getMode()

    /**
     * This method is called to inform the engine that a the player, p, has
     * clicked on country.
     */
    public synchronized int click(String c, Player p) {
        // reject anything not from the current player
        if (players[turn] != p) {
            return NOT_CURRENT_PLAYER;
        }

        // get the country as a country
        Country country = (Country) board.get(c);

        if (country == null) {
            client.error("country " + c + " does not exist");
            return NO_SUCH_COUNTRY;
        }

        // choose by mode the proper response
        switch (mode) {
            case INIT_SELECTION :
                claim(country);
                return OK;

            case INIT_PLACEMENT :
                return initPlace(country);

            case LIMITED_PLACEMENT :
            case FORTIFY_PLACEMENT :
            case PLACEMENT :
                return place(country);

            case ATTACK :
                // if the player owns the contry then selection, else attack
                if (country.getOwner().equals(players[turn])) {
                    return select(country);
                }
                return attack(country);

            case FORTIFY :
                if (country.getOwner().equals(players[turn])) {
                    return fortify(country);
                }
                client.error(Protocol.wNotAllowed);
                return NOT_ALLOWED;

            default :
                throw new InternalServerException("Internal server error (mode is not consistant)");
        }
    } // click( String, Player )

    /**
     * Claim the country as the players own.
     */
    private void claim(Country country) {
        // check the country is not already taken
        if (country.getOwner() != null) {
            client.error("that country is already taken");
            return;
        }

        // claim the country
        country.setOwner(players[turn]);
        country.setArmies(0);

        // select the next player
        do {
            turn++;
            if (turn >= players.length)
                turn = 0;
        } while (!players[turn].playing);

        // decrease the number of unassigned countries
        unassigned--;

        // move to next mode if required
        if (unassigned <= 0) {
            // change mode
            mode = INIT_PLACEMENT;

            // set the initial number of armies
            unassigned = initialArmies / players.length;

            // turn will roll around when we call nextTurn
            turn = players.length;

            // do the next bit
            nextTurn();
        } else {
            // set the client
            client.setTurn(players[turn]);
        }

    } // claim( country )

    /**
     * This method selects the country.
     */
    private int select(Country country) {
        if (country == null) {
            client.select(null);
        } else {
            client.select(country.name);
        }
        selection = country;
        return OK;
    } // select( Country )

    /**
     * This methods is the currently selected player to attack the supplied
     * using the currently selected country.
     */
    private int attack(Country country) {
        // is this sensible?
        if (selection == null) {
            client.error(Protocol.wNoCountry);
        }

        if (!adj.linked(selection.name, country.name)) {
            client.println(RiskServer.warning + Protocol.wNotAdjacent);
            return NOT_ADJACENT;
        }

        // make the best attack possible
        die.bestAttack(selection, country);

        // check for a change of ownership
        if (country.getArmies() < 0) {
            // store the defeated enemy
            Player foe = country.getOwner();

            // draw the country
            country.setOwner(players[turn]);

            // is the foe forever vanquished
            if (foe.playing) {
                if (!isPlaying(foe)) {
                    // tell the client that we have finished with this scum bag
                    client.println(RiskServer.message + foe.name + Protocol.gVanquished);

                    // has the current player won yet?
                    boolean won = true;
                    for (int i = 0; i < players.length; i++) {
                        if (players[i].playing && (i != turn)) {
                            won = false;
                            break;
                        }
                    }
                    if (won) {
                        client.gameOver(players[turn]);
                        return GAME_OVER;
                    }
                }
            }

            // move armies onwards
            armiesInTransit = selection.getArmies() - 1;
            selection.setArmies(0);

            // move an army into the conqued country if possible
            if (armiesInTransit >= 1) {
                armiesInTransit--;
                country.setArmies(1);
            } else {
                country.setArmies(0);
            }

            // set up the status bar
            client.setArmies(RiskStatusBar.status, armiesInTransit);

            // move into limited placement mode if required
            defeated = country;
            if (armiesInTransit > 0) {
                mode = LIMITED_PLACEMENT;
            } else {
                mode = ATTACK;
                select(defeated);
            }

        }

        return OK;
    } // attack( Country )

    /**
     * Test whether a player is still in the game.
     *
     * @return false if the player is already marked as not playing or they do not
     * own any countries. true otherwise.
     */
    private boolean isPlaying(Player p) {
        // simple test
        if (p.playing == false)
            return false;

        // complex test
        for (Enumeration e = board.elements(); e.hasMoreElements();) {
            // if the player still is a country on the board then they are playing
            if (p.equals(((Country) e.nextElement()).getOwner()))
                return true;
        }

        // otherwise the player is not playing
        p.playing = false;
        return false;
    } // isPlayering( Player)

    /**
     * This method is the currently selected player moving an army to the
     * supplied country from the currently selected country.
     */
    private int fortify(Country country) {
        // calculate the number of available armies. The plus one is for army
        // rippling (each 'empty' country contains one army)
        int available = country.getArmies() - country.newArmies + 1;
        available = Math.min(country.getArmies(), available);

        if (available > 0) {
            select(country);
            country.setArmies(country.getArmies() - available);
            client.setArmies(RiskStatusBar.status, available);
            armiesInTransit = available;
            mode = FORTIFY_PLACEMENT;
            return OK;
        } else {
            return NO_ARMIES;
        }
    } // fortify( Country )

    /**
     * This method is the currently selected player placing an army into the
     * supplied country.
     */
    private int initPlace(Country country) {
        // sanity check
        if (armiesInTransit <= 0) {
            throw new InternalServerException("Internal server error (mode is not consistant)");
        }

        // does the player own the country
        if (!players[turn].equals(country.getOwner())) {
            // err ... no
            client.error(Protocol.wNotOwner);
            return NOT_OWNER;
        }

        // change and report state
        client.setArmies(RiskStatusBar.status, (--armiesInTransit));
        country.setArmies(country.getArmies() + 1);

        // move to next player
        if (armiesInTransit <= 0) {
            nextTurn();
        }

        return OK;
    } // initPlace( Country )

    /**
     * This method is the currently selected player placing an army into the
     * supplied country.
     */
    private int place(Country country) {
        // sanity check
        if (armiesInTransit <= 0) {
            throw new InternalServerException("Internal server error (mode is not consistant)");
        }

        // check that the contraints for this type of placement have been met
        switch (mode) {
            case FORTIFY_PLACEMENT :
                // check we are moving armies into an adjacent country
                if (!adj.linked(country.name, selection.name)) {
                    client.error(Protocol.wNotAllowed);
                    return NOT_ALLOWED;
                }
                break;

            case LIMITED_PLACEMENT :
                // check the country is one of its two permited values
                if (!(country.equals(selection) || country.equals(defeated))) {
                    client.error(Protocol.wNotAllowed);
                    return NOT_ALLOWED;
                }
                break;
        }

        // does the player own the country
        if (!players[turn].equals(country.getOwner())) {
            // err ... no
            client.error(Protocol.wNotOwner);
            return NOT_OWNER;
        }

        // change and report state
        client.setArmies(RiskStatusBar.status, (--armiesInTransit));
        country.setArmies(country.getArmies() + 1);
        country.newArmies++;

        // change mode if required
        if (armiesInTransit <= 0) {
            // housekeeping and new mode
            if (mode == FORTIFY_PLACEMENT) {
                mode = FORTIFY;
                select(null);
            } else {
                mode = ATTACK;
                select(country);
            }

        }

        return OK;
    } // place( Country )

    /**
     * Force the turn to move to the next player. This should only ever be
     * called if a player has resigned from the game.
     */
    public synchronized void forceNextTurn() {
        if (mode == INIT_SELECTION) {
            // select the next player
            do {
                turn++;
                if (turn >= players.length)
                    turn = 0;
            } while (!players[turn].playing);

            // let the client know
            client.setTurn(players[turn]);
        } else {
            // this is not actually as hard as I expected!
            armiesInTransit = 0;
            this.nextTurn();
        }
    } // forceNextTurn()

    /**
     * Move to the next player.
     */
    private Player nextTurn() {
        if (mode == INIT_PLACEMENT) {
            // move to the next player
            do {
                turn++;
                if (turn >= players.length) {
                    // we must start again
                    turn = 0;
                    unassigned -= (int) Math.ceil(unassigned / armiesDivisor);
                }
            } while (!players[turn].playing);

            armiesInTransit = (int) Math.ceil(unassigned / armiesDivisor);

            if (unassigned > 0) {
                // notify the client
                client.setTurn(players[turn]);
                client.setArmies(RiskStatusBar.status, armiesInTransit);

                // **** NOTE THIS (important program flow) **** 
                return players[turn];

            } else {
                // finished placing armies, time to start the game

                // this will roll round below
                turn = players.length;
            }
        }

        // set mode
        mode = PLACEMENT;

        // clear the selection
        selection = null;

        // increase turn until a valid player is found
        do {
            turn++;
            if (turn >= players.length)
                turn = 0;
        } while (!players[turn].playing);

        // calculate the number of armies to be awarded (note int division)
        armiesInTransit = Math.max(3, getCountries(players[turn]) / 3);

        // check for continents (does the current player own it

        // OPTIMIZE?
        // (must process continents in constructor to more effiecent storage)
        for (Enumeration e = continents.elements(); e.hasMoreElements();) {
            boolean owns = true;

            Continent c = (Continent) e.nextElement();

            for (Enumeration ce = c.countries(); ce.hasMoreElements();) {
                String name = (String) ce.nextElement();
                Player owner = ((Country) board.get(name)).getOwner();

                if (!owner.equals(players[turn])) {
                    owns = false;

                    // leave the inner loop, player does not own this continent
                    break;
                }
            }

            if (owns) {
                armiesInTransit += c.value;
            }
        }

        // tell the clients
        client.setTurn(players[turn]);
        client.setArmies(RiskStatusBar.status, armiesInTransit);

        return players[turn];
    } // nextTurn()

    /**
     * Count the number of countries a player owns.
     */
    private int getCountries(Player p) {
        // OPTIMIZE?
        int count = 0;

        for (Enumeration e = board.elements(); e.hasMoreElements();) {
            if (p.equals(((Country) e.nextElement()).getOwner())) {
                count++;
            }
        }

        return count;
    }

    /**
     * Player p wants to fortify.
     */
    public synchronized int fortify(Player p) {
        if (!players[turn].equals(p)) {
            return NOT_CURRENT_PLAYER;
        }

        if (!(mode == ATTACK)) {
            client.error(Protocol.wNotAllowed);
            return ILLEGAL_ACTION;
        }

        // clear newarmies for each country
        for (Enumeration e = board.elements(); e.hasMoreElements();) {
            ((Country) e.nextElement()).newArmies = 0;
        }

        // make change visible by getting rid of selection
        select(null);

        mode = FORTIFY;
        return OK;
    } // fortify( Player )

    /**
     * Player p wants to finish their turn.
     */
    public synchronized int finish(Player p) {
        // protect against other players
        if (!players[turn].equals(p)) {
            return NOT_CURRENT_PLAYER;
        }

        // is this permitted
        if (!((mode == ATTACK) || (mode == FORTIFY))) {
            client.error(Protocol.wNotAllowed);
            return ILLEGAL_ACTION;
        }

        nextTurn();

        return OK;
    }

    /**
     * Player p wants to cheat.
     * We can accomadate this by assigned all countries except on to our cheater.
     * @param p
     * @return int
     */
    public synchronized int cheat(Player p) {
    	System.out.println("TODO: cheating is not yet implemented\n");
    	
        return OK;
    }

    /**
     * Data type to store the data associated with a country.
     * <p>
     * This is not a pure data type as the set methods have sideeffects. They
     * broadcast changes in state to the client.
     */
    class Country {
        public final String name;
        // used for fortification
        public int newArmies;

        private int armies = 0;
        private Player owner = null;

        public Country(String n) {
            name = n;
        }

        public final void setOwner(Player p) {
            owner = p;
            client.setOwner(this.name, p);
        }

        public final Player getOwner() {
            return owner;
        }

        public final void setArmies(int a) {
            armies = a;
            client.setArmies(this.name, a);
        }

        public final int getArmies() {
            return armies;
        }

    } // Country.class

} // RiskEngine.class
