// RiskServer.java

package uk.org.redfelineninja.risk.server;

import uk.org.redfelineninja.risk.net.Protocol;
import uk.org.redfelineninja.risk.data.Player;

import uk.org.redfelineninja.risk.io.MapDefinition;
import uk.org.redfelineninja.net.MulticastWriter;
import uk.org.redfelineninja.util.CommandTokenizer;

import java.net.*;
import java.io.*;
import java.util.*;

/**
 * This is a application based server for the Risk Arena.
 *
 * It is capable of running within a Thread to allow it to be easily
 * integrated into the client. It also provides a <code>main</code> method to
 * allow seperate testing.
 */
public class RiskServer extends Thread {
    /**
     * Default port for the Risk server. For a period of time this is the
     * authors favourite number. It has no known historical significance.
     * Try putting 16.44 of petrol in your car so you can enjoy getting
     * a credit card bill. As is so often the way the author has now started
     * flirting with other numbers. So it goes ...
     */
    public static final int defaultPort = 1644;

    /**
     * This is the original risk map (included in the distribution).
     */
    public static final String defaultMap = "maps" + File.separator + "risk.map";

    private String mapFilename;
    private int port;
    private int users = 0;
    private static int serverNum = 0;

    private PrintWriter local;
    private PrintWriter network;
    private MulticastWriter multicaster;

    // these data members will be global to the ServerSession's.
    // To make a multigame server this must be passed arround more carefully
    private RiskEngine engine = null;

    // table of players together with their output streams (talkers and cards)
    private Hashtable playerRegistry;

    // bug work arround ( string concat causes chars to turn to ints)
    // package access
    static final String warning = Protocol.sPrompt + String.valueOf(Protocol.warning);
    static final String message = Protocol.sPrompt + " ";
    static final String comment = Protocol.sPrompt + String.valueOf(Protocol.comment);

    private boolean privacy = false;
    private boolean randomise = true;

    // CONSTRUCTORS

    /**
     * Create a server on the default port using the standard output as the
     * server console. 
     */
    public RiskServer() {
        this(defaultPort, new PrintWriter(System.out));
    }

    /**
     * Start the server on the supplied port using the standard output as the
     * server console.
     */
    public RiskServer(int p) {
        this(p, new PrintWriter(System.out));
    }

    /**
     * Start the sever on the default port using the supplied PrintWriter as
     * the server console.
     */
    public RiskServer(PrintWriter w) {
        this(defaultPort, w);
    }

    /**
     * Start the sever on the supplied port using the supplied PrintWriter as
     * the server console.
     */
    public RiskServer(int p, PrintWriter w) {
        // become a thread properly
        super();

        // simple init
        port = p;
        users = 0;
        local = w;

        // set the map to the default
        mapFilename = defaultMap;

        multicaster = new MulticastWriter();

        // network is a printwriter without automatic flushing
        network = new PrintWriter(multicaster);

        // all server output is displayed on the local display
        multicaster.addListener(local);

        playerRegistry = new Hashtable();
    }

    /**
     * Start the server non-threaded. This is the command to start a single
     * game server.
     */
    public static void main(String args[]) {
        int myPort = defaultPort;
        boolean block = true;

        try {
            for (int i = 0; i < args.length; i++) {
                String arg = args[i].substring(2);

                if (arg.equals("port")) {
                    myPort = Integer.parseInt(args[++i]);
                } else if (arg.equals("noblock")) {
                    block = false;
                } else {
                    System.err.println(
                        "Bad command line arguments\nUsage: RiskServer [--port <port>]");
                    System.exit(1);
                }
            } // for( every argument )
        } catch (Exception e) {
            System.err.println(
                "Bad command line arguments\nUsage: Robot [--port <port>]");
            System.exit(1);
        }

        // this server is NOT run in a seperate thread
        RiskServer rs = new RiskServer(myPort);
        if (block) {
            rs.run();
        } else {
            rs.start();
        }
    }

    // SERVER INTERNALS

    /**
     * This method starts the server and waits for connections. Currently it will
     * never return.
     */
    public void run() {
        ServerSocket server = null;

        try {
            server = new ServerSocket(port);
        } catch (IOException e) {
            local.println("Can not start server on port " + port + ": " + e.getMessage());
            local.flush();
            return;
        }

        // we have started
        local.println("The Risk Arena Server");
        local.println("Copyright (c) 1999 Daniel Thompson and The University of Warwick");
        local.println();
        local.println("Open for connections on " + port);
        local.flush();

        // accept connection 'forever'
        while (true) {
            try {
                // accept a connection
                Socket connection = server.accept();

                // start a new thread to handle the connection
                new ServerSession(connection).start();
            } catch (IOException e) {
                // dump any error info to the servers standard display
                e.printStackTrace(local);
            }
        } // forever
    } // run()

    /**
     * The server session managers each individial client. Each action they
     * make is parsed by this class and the appropriate action taken.
     * <p>
     * 	
     */
    private class ServerSession extends Thread {
        private Socket connection;
        private CommandTokenizer input;
        private PrintWriter output;
        private Player player = null;

        // used to identify a robot
        private boolean robot = false;

        public ServerSession(Socket c) {
            // Thread maintainance
            super();
            setDaemon(true);

            connection = c;
        }

        public void run() {
            // send report to server console
            String host = connection.getInetAddress().getHostName();
            local.println("Accepted connection from " + host);
            local.flush();

            // connect up the input and output streams
            try {
                input = new CommandTokenizer(new InputStreamReader(connection.getInputStream()));
                input.eolIsSignificant(true);
                // this would usually inicate a comment, so make it an ordinary char
                input.ordinaryChar('/');

                output = new PrintWriter(connection.getOutputStream());
            } catch (IOException e) {
                e.printStackTrace(local);
                return;
            }

            // connect out the multicast output from the server
            multicaster.addListener(output);

            // manage the session
            handleSession();

            // exit cleanly (if possible)
            multicaster.removeListener(output);

            // leave the engine is as nice a state as we can
            if (player != null) {

                // this player is no longer playing
                player.playing = false;

                // is there a game running
                if (engine != null) {
                    if (engine.liveGame()) {
                        if (player.equals(engine.getActivePlayer())) {
                            engine.forceNextTurn();
                        }
                    } else {
                        network.println(warning + Protocol.wTerminated);
                        engine = null;
                    } // the game is still alive
                } // there is an engine running

                // remove the player from the system
                playerRegistry.remove(player);
            } // if the player is not null

            try {
                connection.close();
                local.println("Closing connection from " + host);
            } catch (IOException e) {
                local.println("Aborting connection from " + host + " could not close cleanly");
            }

            local.flush();
        } // run()

        private boolean handleInitialisation() throws IOException {
            // this is the template for the expected input
            Character c = input.getCharacter();
            String cmd = input.getWord();

            String name = input.getText();
            Double color = input.getNumber();

            // any nulls and we have the wrong command
            if ((c == null) || (cmd == null) || (name == null) || (color == null)) {
                output.println(message + "  Your first action must be to identify yourself");
                output.flush();
                return false;
            }

            // chech the command
            if (!cmd.equals(Protocol.sIdentify)) {
                output.println(message + "  Your first action must be to identify yourself");
                output.flush();
                return false;
            }

            // store the players details (not initially playing
            player = new Player(name, color.intValue(), false);
            users++;

            // do we have a player of that name already?
            if (playerRegistry.containsKey(player)) {
                output.println(message + "  That name is in use, sorry");
                output.flush();
                player = null;

                return false;
            }

            playerRegistry.put(player, output);

            network.println(player + " joined.");
            network.flush();

            return true;
        }

        private void handleSession() {
            try {
                // get the name of the player (and close the connection if none is
                // given)
                if (!handleInitialisation()) {
                    return;
                }

                // main token taking loop
                main_session : do {
                    input.nextToken();

                    switch (input.ttype) {
                        case Protocol.command :
                            // this is a risk command
                            handleRiskCommand();
                            break;

                        case Protocol.chatCommand :
                            // this is a server command
                            handleServerCommand();
                            break;
                        default :
                            // this is a normal chat command
                            String l = input.getLine();
                            if (l != null) {
                                // allow loopback commands (for testing purpopses only
                                if (l.length() < 2) {
                                    network.println("-" + player + "- " + l);
                                } else if (!l.substring(0, 2).equals(Protocol.sPrompt)) {
                                    network.println("-" + player + "- " + l);
                                } else {
                                    network.println(l);
                                }
                            }
                    } // switch on token type

                    // flush the output
                    network.flush();

                    // clear out any extra tokens on the line
                    while ((input.ttype != input.TT_EOL) && (input.ttype != input.TT_EOF)) {
                        input.nextToken();
                    }
                }
                while (!input.getEOF());
            } catch (IOException e) {
                local.println("Connection lost (can not read input)");
            } catch (NullPointerException e) {
                e.printStackTrace(local);
            }

        } // handleSession()

        private void handleRiskCommand() throws IOException {
            try {
                // all commands are single words
                String cmd = input.getWord();

                // make sure we have a command
                if (cmd == null) {
                    output.println(message + "Commands are always a single word followed by arguments");
                    return;
                }

                // select on command
                if (cmd.equals(Protocol.sStartGame)) {
                    // start the game
                    doStartGame();
                } else if (cmd.equals(Protocol.sCountryClicked)) {
                    // country clicked on
                    doClick();
                } else if (cmd.equals(Protocol.sFortify)) {
                    // request to start fortification
                    doFortify();
                } else if (cmd.equals(Protocol.sFinishTurn)) {
                    // request to finish turn
                    doFinish();
                } else if (cmd.equals(Protocol.sStopGame)) {
                    // tell the connection manager
                    String reason = input.getText();
                } else if (cmd.equals(Protocol.sNewGame)) {
                    // get the game options (random and privacy)
                    doNewGame();
                } else if (cmd.equals(Protocol.sJoinGame)) {
                    // ignored in this version of the server (implict in logging in)
                } else if (cmd.equals(Protocol.sCheat)) {
                    doCheat();
                } else {
                    // not a valid command
                    output.println(comment + Protocol.wInvalidCommand);
                }
            } catch (RuntimeException e) {
                // report the error
                // for robustness a runtime problem should not kill the server
                network.println(warning + Protocol.wServerError);
                local.println(Protocol.gBugReport);
                e.printStackTrace(local);
            }
        } // handleRiskCommand()

        private void handleServerCommand() throws IOException {
            try {
                String command = input.getWord();
                if (command.equals(Protocol.sWho)) {
                    // display the users currently logged in
                    output.println(comment + Protocol.rLoggedOn);

                    for (Enumeration e = playerRegistry.keys(); e.hasMoreElements();) {
                        output.println("        " + e.nextElement().toString());
                    } // for
                } else if (command.equals(Protocol.sRobot)) {
                    // turn the robot on or off
                    command = input.getWord();
                    robot = command.equals(Protocol.sOn);
                    output.println(comment + Protocol.rRobotHints + (robot ? Protocol.sOn : Protocol.sOff));
                } else {
                    output.println(comment + Protocol.rBadCommand);
                }
            } catch (NullPointerException e) {
                // this is caused by a missing argument
                output.println(comment + Protocol.rBadArguments);
            }
        } // handleServerCommand()

        /**
         * Process the input as a request to define a new game.
         */
        private void doNewGame() throws IOException {
            // the name can be ignored since this is a single game server
            String name = input.getText();

            // game flags
            randomise = false;
            privacy = false;

            // get arguments
            String arg;
            try {
                do {
                    arg = input.getWord();
                    if (arg.equals(Protocol.aRandom)) {
                        randomise = true;
                    } else if (arg.equals(Protocol.aPrivacy)) {
                        privacy = true;
                    } else {
                        output.println(message + "invalid argument");
                        return;
                    }
                } while (arg != null);
            } catch (NullPointerException e) {
                // do nothing (end of input)
            }

        } // doNewGame()

        /**
         * Process the input as a request to start a new game.
         */
        private void doStartGame() throws IOException {
            // local must not be used in this synch block
            // this block protects against two users starting the game at the
            // same time
            synchronized (local) {
                // protect against two games on one server
                if (engine != null) {
                    output.println(warning + Protocol.wAlreadyRunning);
                    return;
                }

                // make a new Client writer
                RiskClientWriter client = new RiskClientWriter(network, (Hashtable) playerRegistry.clone());

                // make the players array
                Player[] players = new Player[playerRegistry.size()];
                Enumeration e = playerRegistry.keys();
                for (int i = 0; i < players.length; i++) {
                    // make up the players array for the engine
                    players[i] = (Player) e.nextElement();

                    // make all players registered as such
                    players[i].playing = true;

                    // register all players with the clients
                    client.newPlayer(players[i]);
                }

                // read file
                File map = new File(mapFilename);
                MapDefinition md = new MapDefinition(map);

                // start up the engine
                engine = new RiskEngine(players, md.getAdjacency(), md.getContinents(), client);

                // randomise if requested
                if (randomise)
                    engine.randomise();

                network.println(warning + Protocol.wGameStarted);
            } // synchronized
        } // doStartGame()

        /**
         * Process the input as a click ( # click &lt;country&gt; )
         */
        private void doClick() throws IOException {
            if (engine == null) {
                // use a standard warning (this can be caused by user actions)
                output.println(warning + Protocol.wNoGame);
                return;
            }

            String country = input.getText();

            if (country == null) {
                // this is caused only by a bug
                output.println(message + "must supply country");
                return;
            }

            if (!input.getEOL()) {
                // again this is only caused by a bug in the client
                output.println(message + "too many arguments");
                return;
            }

            int err;
            err = engine.click(country, player);

            // this error would cause a silent failure
            switch (err) {
                case RiskEngine.NOT_CURRENT_PLAYER :
                    output.println(warning + Protocol.wOutOfTurn);
                    break;

                case RiskEngine.GAME_OVER :
                    engine = null;
                    break;

                default :
                    // return a robot hint if it is required
                    if (robot) {
                        output.println(comment + " #" + engine.getMode());
                    }

            } // switch on error

        } // doClick()

        /**
         * Process input as a fortify action ( # fortify )
         */
        private void doFortify() throws IOException {
            if (engine == null) {
                // use a standard warning (this can be caused by user actions)
                output.println(warning + Protocol.wNoGame);
                return;
            }

            if (!input.getEOL()) {
                // this is only caused by a bug in the client
                output.println(message + "too many arguments");
                return;
            }

            // check for potentially silent error (and voice it)
            if (engine.fortify(player) == RiskEngine.NOT_CURRENT_PLAYER) {
                output.println(warning + Protocol.wOutOfTurn);
            }

            // return a robot hint if required
            if (robot) {
                output.println(comment + " #" + engine.getMode());
            }
        } // doFortify()

        /**
         * Process input as a finish action ( # finish )
         */
        private void doFinish() throws IOException {
            if (engine == null) {
                // use a standard warning (this can be caused by user actions)
                output.println(warning + Protocol.wNoGame);
                return;
            }

            if (!input.getEOL()) {
                // this is only caused by a bug in the client
                output.println(message + "too many arguments");
                return;
            }

            // check for potentially silent error (and voice it)
            if (engine.cheat(player) == RiskEngine.NOT_CURRENT_PLAYER) {
                output.println(warning + Protocol.wOutOfTurn);
            }

            // return a robot hint if required
            if (robot) {
                output.println(comment + " #" + engine.getMode());
            }
        }

        /**
         * Alter the risk map so that the current player owns all of the
         * countries except one.
         * This is only really intended as a debug option and should really
         * be protected from use by a real player.
         */
        private void doCheat() throws IOException {
            if (engine == null) {
                // use a standard warning (this can be caused by user actions)
                output.println(warning + Protocol.wNoGame);
                return;
            }

            if (!input.getEOL()) {
                // this is only caused by a bug in the client
                output.println(message + "too many arguments");
                return;
            }

            // check for potentially silent error (and voice it)
            if (engine.finish(player) == RiskEngine.NOT_CURRENT_PLAYER) {
                output.println(warning + Protocol.wOutOfTurn);
            } else {
                output.println(comment + "All lands (bar one) have given to a cheat\n");
            }
            
            // return a robot hint if required
            if (robot) {
                output.println(comment + " #" + engine.getMode());
            }
        }


    } // ServerSession.class

} // RiskServer.class
