// Adjacency.java

package uk.org.redfelineninja.util;

import java.util.*;

/**
 * A data structure to store Adjanceny properties.
 * <p>
 * This data structure does not have strict safety contraints. Errors are passed
 * the the programmer as exceptions from the underlying code so their
 * interpretation must be with care.
 *
 * @author	Daniel Thompson, Sep 1998
 * @version	0.5
 */
public class Adjacency {

    // Data Members

    /**
     * Stores the labels, usually these are strings
     */
    private Vector labels;

    /**
     * Stores the actual status information
     */
    private boolean matrix[][];

    private int capacity;

    // Methods

    /**
     * Creates a new empty Adjancency matrix.
     */
    public Adjacency() {
        labels = new Vector();
        capacity = labels.capacity();
        matrix = new boolean[capacity][capacity];
    }

    /**
     * Creates a new Adjancency matrix.
     *
     * @param	v	A vector to supply the headings
     */
    public Adjacency(Vector v) {
        // this must be cloned as we do not want it to be changed
        labels = (Vector) v.clone();
        capacity = labels.capacity();
        matrix = new boolean[capacity][capacity];
    }

    /**
     * Makes a link from a to b
     *
     * @param	a	Source label
     * @param	b	Dest. label
     */
    public void link(Object a, Object b) {
        matrix[labels.indexOf(a)][labels.indexOf(b)] = true;
    }

    /**
     * Makes a two directional link between a and b
     *
     * @param	a	label
     * @param	b	label
     */
    public void doubleLink(Object a, Object b) {
        matrix[labels.indexOf(a)][labels.indexOf(b)] = true;
        matrix[labels.indexOf(b)][labels.indexOf(a)] = true;
    }

    /**
    * Ensures that there is no link from a to b
    *
    * @param	a	Source label
    * @param	b	Dest. label
    */
    public void unlink(Object a, Object b) {
        matrix[labels.indexOf(a)][labels.indexOf(b)] = false;
    }

    /**
     * Ensures that there is no link in either direction between a and b
     *
     * @param	a	label
     * @param	b	label
     */
    public void doubleUnlink(Object a, Object b) {
        matrix[labels.indexOf(a)][labels.indexOf(b)] = false;
        matrix[labels.indexOf(b)][labels.indexOf(a)] = false;
    }

    /**
     * Tests for a link from a to b
     *
     * @param	a	Source label
     * @param	b	Dest. label
     */
    public boolean linked(Object a, Object b) {
        return matrix[labels.indexOf(a)][labels.indexOf(b)];
    }

    /**
     * Add another label to the data structure.
     * If required the data structure will resize itself.
     *
     * @param	o	Label to be added
     */
    public void addElement(Object o) {
        labels.addElement(o);

        // if capcity has changed
        if (labels.capacity() != capacity) {
            // create a new matrix
            boolean newMatrix[][] = new boolean[labels.capacity()][labels.capacity()];

            // copy the matrix
            for (int i = 0; i < capacity; i++)
                for (int j = 0; j < capacity; j++)
                    newMatrix[i][j] = matrix[i][j];

            // update
            matrix = newMatrix;

        } // if

        // we will set this every time for safety
        capacity = labels.capacity();

    } // addElement

    public String toString() {
        StringBuffer sb = new StringBuffer();
        Enumeration e;

        sb.append("  ");
        e = labels.elements();
        while (e.hasMoreElements()) {
            sb.append(e.nextElement().toString().substring(0, 1).toUpperCase()).append(" ");
        }
        sb.append("\n");

        e = labels.elements();
        for (int i = 0; i < labels.size(); i++) {
            sb.append("  ");

            for (int j = 0; j < labels.size(); j++) {
                sb.append(matrix[i][j] ? "* " : ". ");
            }

            sb.append("   ").append(e.nextElement()).append("\n");
        }

        return sb.toString();
    }

    // MISC VECTOR OPERATIONS

    /**
     * Tests if the specifed object is a member of the Adjaceny matrix.
     */
    public boolean contains(Object obj) {
        return labels.contains(obj);
    }

    /**
     * Gives the size of this adjacency matrix. Since the matrix must be square
     * this is the length of one side.
     */
    public int size() {
        return labels.size();
    }

    /**
     * Supplies an enumeration of all the labels used in this matrix.
     */
    public Enumeration labels() {
        return labels.elements();
    }

    /**
     * Copies the labels of this matrix into an array. The array must be long
     * enough to contain all the labels.
     */
    public void copyLabelsInto(Object[] objs) {
        labels.copyInto(objs);
    }
} // Adjacency.class
