// CommandTokenizer.java

/*
 * Copyright (c) 1999 Daniel Thompson.
 *
 * This software is licenced under the artistic licence. You may use this
 * software in anything but my name must remain as the original author.
 *
 * This software is supplied as is with no garantees either express or implied.
 *
 * This software was writen clean room (no references were made to the Java
 * source code).
 */

package uk.org.redfelineninja.util;

import java.io.*;

/**
 * This class provides a simple means to parse strings in command driven
 * situations. This class is almost equivalent to java.io.StreamTokenizer but
 * can be used in command driven situations. java.io.StreamTokenizer can not
 * detect the end of a line until the next line has been presented to it. This
 * is not approptiate in command driven situations. This class can be used as
 * a drop in replacement in a command driven system or as an advanded version
 * of java.util.StringTokenizer.
 * <p>
 * WARNING: This has not yet acheived its goal as a drop in replacement for
 * StreamTokenizer. It has no decent comment handling.
 *
 * @author Daniel Thompson, Feb 1999
 * @version 0.8
 */
public class CommandTokenizer {

    /**
     * The string value of the current token if applicable.
     */
    public String sval;

    /**
     * The numeric value of the current token if applicable.
     */
    public double nval;

    /** 
     * The type of the current token.
     */
    public int ttype = TT_UNDEF;

    // this is to store ttype when pushed back
    private int ttypeold;

    /**
     * ttype takes this value if the end of the file has been reached.
     */
    public static final int TT_EOF = -1;

    /**
     * ttype takes this value if the current token is numeric.
     */
    public static final int TT_NUMBER = -2;

    /**
     * ttype takes this value if the current token is a single word.
     */
    public static final int TT_WORD = -3;

    /**
     * ttype takes this value at the end of a line. This value will not be
     * returned unless EOLs are made significant.
     *
     * @see CommandTokenizer#eolIsSignificant
     */
    public static final int TT_EOL = '\n';

    /**
     * ttype takes this value when pushed back and at any other time when its
     * value is not defined. This can generally be ignored as a possibility in
     * well writen code.
     */
    public static final int TT_UNDEF = 0;

    // PRIVATE STATE VARIABLES
    private int[] normalChar = { 'A', 'z' };
    private final int decimal = '.';
    private char[] whiteSpace = { '	', ' ' };
    private char[] stringDelim = { '\'', '"' };
    private char[] comment = { '/' };

    private boolean pushedBack = false;
    private boolean eolSignificant = false;

    // THESE ARE THE GUTS
    private BufferedReader reader = null;
    private String command = null;
    private int position;

    /**
     * Create a new CommandTokenizer that uses the supplied InputStream.
     *
     * Refer to CommandTokenizer( Reader ) for default starting states.
     *
     * @deprecated The prefered way to create a CommandTokenizer is by supplying
     * it with a Reader.
     * @see CommandTokenizer#( Reader )
     * @see java.io.Reader
     */
    public CommandTokenizer(InputStream is) {
        this(new BufferedReader(new InputStreamReader(is)));
    }

    /**
     * Create a new CommandTokenizer that uses the supplied Reader.
     *
     * This default state of the CommandTokenizer is as follows.
     * <p><ul>
     * <li> tab and space codes are white space.
     * <li> " and ' are string deliminators
     * <li> / denotes comments
     * <li> ordinary characters are considered to be in the range 'A' to 'z'
     * <li> numerics are detected and parsed into doubles automatically
     * </ul>
     */
    public CommandTokenizer(Reader r) {
        if (r instanceof BufferedReader) {
            reader = (BufferedReader) r;
        } else {
            reader = new BufferedReader(r);
        }
    }

    /**
     * Create a new CommandTokenizer that takes its data from the supplied
     * string.
     * <p>
     * The CommandTokenizer will behave exactly as though the string was
     * received from a Reader. When all tokens in the string have been
     * exhusted the tokenizer will constantly return TT_EOF.
     */
    public CommandTokenizer(String c) {
        command = c;
        position = 0;
    }

    // ADMIN AND STATE FUNCTIONS

    /**
     * Sets whether or not EOLs will be treated as tokens. Setting this to true
     * will cause the tokenizer to return TT_EOL every time the end of a line is
     * reached.
     */
    public void eolIsSignificant(boolean b) {
        eolSignificant = b;
    }

    public char[] getStringDelim() {
        char[] s = new char[stringDelim.length];

        for (int i = 0; i < stringDelim.length; i++) {
            s[i] = stringDelim[i];
        }

        return s;
    }

    // SYNTAX MODIFICATIONS

    public void commentChar(int ch) {
        comment = addPeer(comment, (char) ch);
    }

    public void ordinaryChar(int ch) {
        removePeer(whiteSpace, (char) ch);
        removePeer(stringDelim, (char) ch);
        removePeer(comment, (char) ch);
    }

    // remove a character from an array (delete inplace using 0 as null
    private static void removePeer(char[] chars, char ch) {
        // this memory leak is made safe by the insert method 
        for (int i = 0; i < chars.length; i++) {
            if (chars[i] == ch)
                chars[i] = 0;
        }
    } // removePeer( ... )

    // add a character to an array (filling in nulls is possible)
    private static char[] addPeer(char[] chars, char ch) {
        int tag = -1;

        char[] newChars = new char[chars.length + 1];

        for (int i = 0; i < chars.length; i++) {
            newChars[i] = chars[i];
            if (newChars[i] == 0)
                tag = i;
        }

        if (tag < 0) {
            newChars[chars.length] = ch;
            return newChars;
        } else {
            chars[tag] = ch;
            return chars;
        }

    } // addPeer( ... )

    // TEXT PROCESSING FUNCTIONS

    /**
     * Pushes the current token back so that nextToken will return it again.
     */
    public synchronized void pushBack() {
        ttypeold = ttype;
        ttype = TT_UNDEF;
        pushedBack = true;
    }

    /**
     * Get the whole of the current line for processing in a different way. 
     * This will move the CommandTokenizer onto the next line. Even it EOL is
     * significant the tokenizer will not supplied and EOL signal. To make
     * things easier for EOL detectors the token type (ttype) is changed to
     * TT_EOL
     */
    public String getLine() {
        String tmp = command;

        // stops stupid things happening
        pushedBack = false;

        // invalidate this line (causes new line to be read)

        // we are now at the end of a line
        ttype = TT_EOL;

        command = null;
        return tmp;
    }

    /**
     * Move the tokenizer to the next token.
     * <p>
     * The tokens are categorized as follows.
     * <p><ul>
     * <li>End of file, return TT_EOF
     * <li>End of line, return TT_EOL (if eol is significant)
     * <li>Word, return TT_WORD
     * <li>Number, return TT_NUMBER
     * <li>String, return the integer value of the string deliminator
     * <li>Special Character, return that character (cast to int)
     * </ul>
     * @return a integer to indentfy the current token
     */
    public synchronized int nextToken() throws IOException {
        // are we pushed back?
        if (pushedBack) {
            ttype = ttypeold;
            pushedBack = false;
            return ttype;
        }

        // default for ttype
        ttype = TT_EOL;

        if (command == null) {
            if (reader == null) {
                // all data consumed
                ttype = TT_EOF;
                return ttype;
            } else {
                // try get another line
                try {
                    command = reader.readLine();

                    if (command == null) {
                        ttype = TT_EOF;
                        return ttype;
                    }
                } catch (EOFException e) {
                    command = null;
                    ttype = TT_EOF;
                    return ttype;
                }

                // if another line was got we are on the first character
                position = 0;
            } // if
        } // if ( command is null )

        // decide the type of this token
        if (position == command.length()) {
            ttype = TT_EOL;
            command = null;
        } else {
            int first = command.charAt(position);

            if ((first >= normalChar[0]) && (first <= normalChar[1])) {
                // word
                ttype = TT_WORD;
                sval = getWordP();
            } else if (isNumericP(command, position)) {
                // number
                ttype = TT_NUMBER;
                nval = getNumberP();
            } else if (in(first, stringDelim)) {
                // string
                ttype = first;
                sval = getStringP(first);
            } else if (in(first, comment)) {
                // comment - ignore the rest of the line
                ttype = TT_EOL;
                command = null;
            } else {
                // special character
                ttype = first;
                sval = String.valueOf((char) ttype);
                position++;
            }
        } // outer if 

        if ((ttype == TT_EOL) && (!eolSignificant)) {
            // if we are supposed to ignore EOLs then get another token
            ttype = nextToken();
        }

        // wind off any white space
        if (command != null) {
            while (position < command.length()) {
                if (in(command.charAt(position), whiteSpace)) {
                    position++;
                } else {
                    break;
                }
            }
        } // kill white space

        return ttype;
    } // nextToken()

    /**
     * Winds on the input to the end of the word.
     *
     * @return the word that has been consumed
     */
    private String getWordP() {
        int start = position;

        int ch = command.charAt(position);

        // copy characters until we reach the end of a word
        while ((ch >= normalChar[0]) && (ch <= normalChar[1])) {
            // next char
            position++;

            // have we reached the end of the line
            if (position < command.length()) {
                ch = command.charAt(position);
            } else {
                ch = TT_EOL;
            }
        }

        return command.substring(start, position);
    }

    /**
     * Read a numeric value from the command.
     * <p>
     * If the input include 1.1.1 this will be treated as two tokens 1.1 and 0.1.
     * Although this is arguably wrong this is the behaviour of StreamTokenizer.
     *
     * @return the number consumed (as a double)
     */
    private double getNumberP() {
        int start = position;
        int ch;

        // wind of numeral before decimal point
        do { // while current character is a numeral
            // next char
            position++;

            // have we reached the end of the line
            if (position < command.length()) {
                ch = command.charAt(position);
            } else {
                ch = TT_EOL;
            }
        } while (Character.isDigit((char) ch));

        if (position < command.length()) {
            // enforce lazy eval.
            if (command.charAt(position) == decimal) {
                // if there is a decimal point wind off some more
                position++;

                // is position too large?
                ch = (position < command.length() ? command.charAt(position) : TT_EOL);

                while (Character.isDigit((char) ch)) {
                    // next char
                    position++;

                    // have we reached the end of the line
                    if (position < command.length()) {
                        ch = command.charAt(position);
                    } else {
                        ch = TT_EOL;
                    }
                } // current character is a numeral
            }
        } // outer if

        // position points to the next valid char (substring is exclusive for end)
        sval = command.substring(start, position);
        return new Double(sval).doubleValue();
    }

    /**
     * Winds on the input until another string deliminator is found. This is
     * more thorough than StreamTokenizer since it can handle EOLs as part of
     * the String.
     *
     * @return the string consumed
     */
    private String getStringP(int delim) throws IOException {
        StringBuffer str = new StringBuffer();
        int end;

        position++;

        do {
            end = command.indexOf(delim, position);

            if (end == -1) {
                // quotes go off end of line
                str.append(command.substring(position) + '\n');

                // no exception handling - EOF IS an error here
                command = reader.readLine();
                position = 0;
            } else {
                // copy until the end of the quotes.
                str.append(command.substring(position, end));

                // set position past quotes
                position = end + 1;
            }
        } while (end == -1);

        return str.toString();
    }

    // UTILITIES

    /**
     * Checks a charecter for membership of the supplied array.
     */
    public static final boolean in(int ch, char[] chars) {
        for (int i = 0; i < chars.length; i++) {
            if (ch == chars[i])
                return true;
        }

        return false;
    } // in( char, char [] )

    /**
     * Test if the Token is numeric.
     *
     * @return true if the token is numeric
     */
    public final boolean isNumeric() {
        return (ttype == TT_NUMBER);
    }

    /**
     * Test if the token is a single word.
     *
     * @return true if the token is a single word
     */
    public final boolean isWord() {
        return (ttype == TT_WORD);
    }

    /**
     * Test if the token is a quoted string.
     *
     * @return true if the token is a quoted string
     */
    public final boolean isString() {
        return in(ttype, stringDelim);
    }

    /**
     * Test if sval is defined for this token.
     *
     * @return true if sval is defined
     */
    public final boolean isText() {
        return isWord() || isString();
    }

    /**
     * Try to get a number from the tokenizer. If we can't get a number
     * then we call pushBack so the token is not consumed.
     *
     * @return a Double if the input was interpreted as a number otherwise null
     */
    public final Double getNumber() throws IOException {
        nextToken();

        if (ttype == TT_NUMBER) {
            return new Double(nval);
        } else {
            pushBack();
            return null;
        }
    }

    /**
     * Try to get a text token from the tokenizer. If we can't get text
     * then we call pushBack so the token is not consumed.
     *
     * @return a String if the input was interpreted as text otherwise null
     */
    public final String getText() throws IOException {
        nextToken();

        if ((ttype == TT_WORD) || in(ttype, stringDelim)) {
            return sval;
        } else {
            pushBack();
            return null;
        }
    }

    /**
     * Try to get a string from the supplied tokenizer. If we can't get a string
     * then we call pushBack so the token is not consumed.
     *
     * @param tok the tokenizer
     * @param quotes the characters set in the tokenizer to denote strings
     *
     * @return a String if the input was interpreted as a string otherwize null
     */
    public final String getString() throws IOException {
        if (in(nextToken(), stringDelim)) {
            return sval;
        } else {
            pushBack();
            return null;
        }
    }

    /**
     * Try to get a word from the tokenizer. If we can't get a word
     * then we call pushBack so the token is not consumed.
     *
     * @param tok the tokenizer
     *
     * @return a String if the input was interpreted as a word otherwize null
     */
    public final String getWord() throws IOException {
        if (nextToken() == TT_WORD) {
            return sval;
        } else {
            pushBack();
            return null;
        }
    }

    /**
     * Try to get a character from the tokenizer. If we can't get a
     * character then we call pushBack so the token is not consumed.
     *
     * @return a Character if the input was interpreted as a character otherwize
     * null
     */
    public final Character getCharacter() throws IOException {
        switch (nextToken()) {
            case TT_WORD :
            case TT_NUMBER :
            case TT_EOL :
            case TT_EOF :
                pushBack();
                return null;

            default :
                if (in(ttype, stringDelim)) {
                    pushBack();
                    return null;
                }

                return new Character((char) ttype);
        }
    }

    /**
     * Test whether the tokenizer has reached the end of a line. It is
     * assumed that if we reach the EOF marker then we have also reached the end
     * of the line. 
     * @see java.io.StreamTokenizer#eolIsSignificant
     * @return true if the end of line or file has been reached, false otherwise
     * and the input is pushed back so the token is not consumed
     */
    public final boolean getEOL() throws IOException {
        nextToken();

        if ((ttype == TT_EOL) || (ttype == TT_EOF)) {
            return true;
        } else {
            pushBack();
            return false;
        }
    }

    /**
     * Test whether the tokenizer has reached the end of a file. If the
     * test fails we do not consume the token.
     *
     * @return true if the end of file has been reached.
     */
    public final boolean getEOF() throws IOException {
        if (nextToken() == TT_EOF) {
            return true;
        } else {
            pushBack();
            return false;
        }
    }

    /**
     * Checks whether the current input is numeric.
     * <p>
     * This would be static but it uses some of the private state variables.
     */
    private final boolean isNumericP(String s, int p) {
        // EOL is not numeric
        if (p >= s.length())
            return false;

        int ch = s.charAt(p);

        // decial or - should be skiped and the next character checked if
        // possible
        if ((p + 1 < s.length()) && ((ch == '-') || (ch == decimal))) {
            ch = s.charAt(p + 1);
        }

        // any digit is numeric
        if (Character.isDigit((char) ch))
            return true;

        // anything else is not numeric
        return false;
    } // isNumeric( String, int )

} // CommandTokenizer.class
