// IntStack.java

package uk.org.redfelineninja.util;

/**
 * This stack class represents a last-in-first-out (LIFO) stack of integers.
 * In some cases speed is of paramount importance, in such cases <code>
 * java.util.Stack</code> should be avoided since its use forces object
 * creation. Object creation is time consuming.
 * <p>
 * The method names used in this class are taken from <code>java.util.Vector
 * </code> and <code>java.util.Stack</code> so there should be little memory
 * burdon in using this class.
 *
 * @author Daniel Thompson (Feb 1999)
 */
public class IntStack {

    // DATA MEMBERS

    /**
     * default size of the array and increment when the array is enlarged.
     */
    protected final int capacityIncrement;

    /**
     * head of the stack.
     */
    protected int head = -1;

    /**
     * the stack.
     */
    protected int[] stack;

    // CONSTRUCTORS

    /**
     * Creates an empty stack with the default initial size.
     */
    public IntStack() {
        capacityIncrement = 512;
        stack = new int[capacityIncrement];

        // jdk1.1 did not like this
        // this( 512 );
    }

    /**
     * Creates an empty stack with specified initial size.
     *
     * @param size initial size and capacity increment
     */
    public IntStack(int size) {
        capacityIncrement = size;
        stack = new int[capacityIncrement];
    }

    // MAIN METHODS

    /**
     * Push an item onto the stack.
     *
     * @param item the item to be pushed
     */
    public final void push(int item) {
        try {
            head++;
            stack[head] = item;
        } catch (ArrayIndexOutOfBoundsException e) {
            // lets be safe, underflow or overflow
            if (head >= stack.length) {
                // make space
                grow();
            } else {
                // stack has been abused, reset
                head = 0;
            }

            // now we can put the data in
            stack[head] = item;
        }
    } // push( int )

    /**
     * Pop an item off the stack.
     *
     * @return the item removed
     */
    public final int pop() {
        // this is a POST decrement (Bounds exception thrown before decrement)
        return stack[head--];
    }

    /**
     * Peek at the top of the stack.
     *
     * @return the value at the top of the stack
     */
    public final int peek() {
        return stack[head];
    }

    /**
     * Test whether or not the stack is empty.
     *
     * @return true if the stack is empty
     */
    public final boolean empty() {
        return (head == -1);
    }

    /**
     * Get the size of the stack.
     *
     * @return size
     */
    public final int size() {
        return head;
    }

    // UTILITY METHODS

    /**
     * Increase the stack size by capacity increment and maintain consistancy.
     */
    protected final void grow() {
        int[] oldStack;

        oldStack = stack;
        stack = new int[oldStack.length + capacityIncrement];

        // limit on length, this MAY occassionally be inefficient but it is SAFE
        // (head cound be in an inconsistant state during a grow)
        for (int i = 0; i < oldStack.length; i++) {
            stack[i] = oldStack[i];
        } // for
    }
} // IntStack.class
