// Token.java

package uk.org.redfelineninja.util;

import java.io.*;

/**
 * This class is a companion to the various Java tokenizers. It is an immutable
 * class that initializes itself from the current state of a tokenizer.
 * <p>
 * It is most tightly intergrated with CommandTokenizer but should work with
 * StreamTokenizer.
 *
 * @see CommandTokenizer
 * @see java.io.StreamTokenizer
 */
public class Token {
    public final int ttype;
    public final double nval;
    public final String sval;
    private final char[] stringDelim;

    public static final int TT_EOF = StreamTokenizer.TT_EOF;
    public static final int TT_EOL = StreamTokenizer.TT_EOL;
    public static final int TT_WORD = StreamTokenizer.TT_WORD;
    public static final int TT_NUMBER = StreamTokenizer.TT_NUMBER;

    /**
     * Create a new Token from a CommandTokenizer. All the methods will
     * work correctly.
     */
    public Token(CommandTokenizer tokenizer) {
        ttype = tokenizer.ttype;
        nval = tokenizer.nval;
        sval = tokenizer.sval;

        stringDelim = tokenizer.getStringDelim();
    }

    /**
     * Create a new Token from a StreamTokenizer. The token will be unable to
     * determine what the string deliminators are, therefore the isString() and
     * isText() methods will throw NullPointerExceptions if called. In addition
     * toString() will no longer return quoted strings, it will simply return
     * a quote character. All information will be preserved.
     */
    public Token(StreamTokenizer tokenizer) {
        ttype = tokenizer.ttype;
        nval = tokenizer.nval;
        sval = tokenizer.sval;

        stringDelim = null;
    }

    /**
     * Create a new Token from a StreamTokenizer. The token will treat the
     * the supplied array of characters are the set of all string deliminators.
     * If that set is correct all the methods will work correctly.
     */
    public Token(StreamTokenizer tokenizer, char[] s) {
        ttype = tokenizer.ttype;
        nval = tokenizer.nval;
        sval = tokenizer.sval;

        stringDelim = s;
    }

    /**
     * Convert the current token to printable form.
     */
    public String toString() {
        if (TT_EOF == ttype) {
            return "EOF";
        } else if (TT_EOL == ttype) {
            return "EOL";
        } else if (TT_WORD == ttype) {
            return sval;
        } else if (TT_NUMBER == ttype) {
            return String.valueOf(nval);
        } else if (null == stringDelim) {
            return String.valueOf((char) ttype);
        } else if (isString()) {
            return sval;
        } else {
            return String.valueOf((char) ttype);
        }
    }

    // TESTS

    /**
     * Test if the Token is numeric.
     *
     * @return true if the token is numeric
     */
    public final boolean isNumeric() {
        return (ttype == TT_NUMBER);
    }

    /**
     * Test if the token is a single word.
     *
     * @return true if the token is a single word
     */
    public final boolean isWord() {
        return (ttype == TT_WORD);
    }

    /**
     * Test if the token is a quoted string.
     *
     * @return true if the token is a quoted string
     */
    public final boolean isString() {
        return CommandTokenizer.in(ttype, stringDelim);
    }

    /**
     * Test if sval is defined for this token.
     *
     * @return true if sval is defined
     */
    public final boolean isText() {
        return isWord() || isString();
    }

    /**
     * Try to get a number from the supplied tokenizer. If we can't get a number
     * then we call pushBack so the token is not consumed.
     *
     * @return a Double if the input was interpreted as a number otherwise null
     */
    public static final Double getNumber(StreamTokenizer tok) throws IOException {
        if (tok.nextToken() == tok.TT_NUMBER) {
            return new Double(tok.nval);
        } else {
            tok.pushBack();
            return null;
        }
    }

    /**
     * Try to get a text token from the supplied tokenizer. If we can't get text
     * then we call pushBack so the token is not consumed.
     *
     * @param tok the tokenizer
     * @param quotes the characters set in the tokenizer to denote strings
     *
     * @return a String if the input was interpreted as text otherwise null
     */
    public static final String getText(StreamTokenizer tok, char[] quotes) throws IOException {
        tok.nextToken();

        if ((tok.ttype == tok.TT_WORD) || CommandTokenizer.in(tok.ttype, quotes)) {
            return tok.sval;
        } else {
            tok.pushBack();
            return null;
        }
    }

    /**
     * Try to get a string from the supplied tokenizer. If we can't get a string
     * then we call pushBack so the token is not consumed.
     *
     * @param tok the tokenizer
     * @param quotes the characters set in the tokenizer to denote strings
     *
     * @return a String if the input was interpreted as a string otherwize null
     */
    public static final String getString(StreamTokenizer tok, char[] quotes) throws IOException {
        if (CommandTokenizer.in(tok.nextToken(), quotes)) {
            return tok.sval;
        } else {
            tok.pushBack();
            return null;
        }
    }

    /**
     * Try to get a word from the supplied tokenizer. If we can't get a word
     * then we call pushBack so the token is not consumed.
     *
     * @param tok the tokenizer
     *
     * @return a String if the input was interpreted as a word otherwize null
     */
    public static final String getWord(StreamTokenizer tok) throws IOException {
        if (tok.nextToken() == tok.TT_WORD) {
            return tok.sval;
        } else {
            tok.pushBack();
            return null;
        }
    }

    /**
     * Try to get a character from the supplied tokenizer. If we can't get a
     * character then we call pushBack so the token is not consumed.
     *
     * @param tok the tokenizer
     * @param quotes the characters set in the tokenizer to denote strings
     *
     * @return a Character if the input was interpreted as a character otherwize
     * null
     */
    public static final Character getCharacter(StreamTokenizer tok, char[] quotes) throws IOException {
        int ttype = tok.nextToken();
        if (TT_WORD == ttype || TT_NUMBER == ttype || TT_EOL == ttype || TT_EOF == ttype) {
            tok.pushBack();
            return null;
        }

        if (CommandTokenizer.in(tok.ttype, quotes)) {
            tok.pushBack();
            return null;
        }

        return new Character((char) tok.ttype);
    }

    /**
     * Test whether the supplied tokenizer has reached the end of a line. It is
     * assumed that if we reach the EOF marker then we have also reached the end
     * of the line. 
     * @see java.io.StreamTokenizer#eolIsSignificant
     * @return true if the end of line or file has been reached, false otherwise
     * and the input is pushed back so the token is not consumed
     */
    public static final boolean getEOL(StreamTokenizer tok) throws IOException {
        tok.nextToken();

        if ((tok.ttype == tok.TT_EOL) || (tok.ttype == tok.TT_EOF)) {
            return true;
        } else {
            tok.pushBack();
            return false;
        }
    }

    /**
     * Test whether the supplied tokenizer has reached the end of a file. If the
     * test fails we do not consume the token.
     *
     * @return true if the end of file has been reached.
     */
    public static final boolean getEOF(StreamTokenizer tok) throws IOException {
        if (tok.nextToken() == tok.TT_EOF) {
            return true;
        } else {
            tok.pushBack();
            return false;
        }
    }
}
